#------------------------------------------------------------------------------
# Copyright (c) 2020, 2022, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

"""
3500 - Module for testing the JSON data type.
"""

import datetime
import decimal
import unittest

import oracledb
import test_env

@unittest.skipUnless(test_env.get_client_version() >= (21, 0),
                     "unsupported client")
@unittest.skipUnless(test_env.get_server_version() >= (21, 0),
                     "unsupported server")
class TestCase(test_env.BaseTestCase):

    json_data = [
        True,
        False,
        'String',
        b'Some Bytes',
        {},
        {"name": None},
        {"name": "John"},
        {"age": 30},
        {"Permanent": True},
        {
            "employee": {
                "name":"John",
                "age": 30,
                "city": "Delhi",
                "Parmanent": True
            }
        },
        {
            "employees": ["John", "Matthew", "James"]
        },
        {
            "employees": [
                {
                    "employee1": {"name": "John", "city": "Delhi"}
                },
                {
                    "employee2": {"name": "Matthew", "city": "Mumbai"}
                },
                {
                    "employee3": {"name": "James", "city": "Bangalore"}
                }
            ]
        }
    ]

    def __bind_scalar_as_json(self, data):
        self.cursor.execute("truncate table TestJson")
        out_var = self.cursor.var(oracledb.DB_TYPE_JSON,
                                  arraysize=len(data))
        self.cursor.setinputsizes(None, oracledb.DB_TYPE_JSON, out_var)
        bind_data = list(enumerate(data))
        self.cursor.executemany("""
                insert into TestJson values (:1, :2)
                returning JsonCol into :json_out""", bind_data)
        self.connection.commit()
        self.assertEqual(out_var.values, [[v] for v in data])

    def test_3500_insert_and_fetch_single_json(self):
        "3500 - insert and fetch single row with JSON"
        self.cursor.execute("truncate table TestJson")
        self.cursor.setinputsizes(None, oracledb.DB_TYPE_JSON)
        self.cursor.execute("insert into TestJson values (:1, :2)",
                            [1, self.json_data])
        self.cursor.execute("select JsonCol from TestJson")
        result, = self.cursor.fetchone()
        self.assertEqual(result, self.json_data)

    def test_3501_execute_with_dml_returning(self):
        "3501 - inserting single rows with JSON and DML returning"
        json_val = self.json_data[11]
        self.cursor.execute("truncate table TestJson")
        json_out = self.cursor.var(oracledb.DB_TYPE_JSON)
        self.cursor.setinputsizes(None, oracledb.DB_TYPE_JSON, json_out)
        self.cursor.execute("""
                insert into TestJson values (:1, :2)
                returning  JsonCol into :json_out""",
                [1, json_val])
        self.assertEqual(json_out.getvalue(0), [json_val])

    def test_3502_insert_and_fetch_multiple_json(self):
        "3502 - insert and fetch multiple rows with JSON"
        self.cursor.execute("truncate table TestJson")
        self.cursor.setinputsizes(None, oracledb.DB_TYPE_JSON)
        data = list(enumerate(self.json_data))
        self.cursor.executemany("insert into TestJson values(:1, :2)", data)
        self.cursor.execute("select * from TestJson")
        fetched_data = self.cursor.fetchall()
        self.assertEqual(fetched_data, data)

    def test_3503_executemany_with_dml_returning(self):
        "3503 - inserting multiple rows with JSON and DML returning"
        self.cursor.execute("truncate table TestJson")
        int_values = [i for i in range(len(self.json_data))]
        out_int_var = self.cursor.var(int, arraysize=len(int_values))
        out_json_var = self.cursor.var(oracledb.DB_TYPE_JSON,
                                       arraysize=len(int_values))
        self.cursor.setinputsizes(None, oracledb.DB_TYPE_JSON, out_int_var,
                                  out_json_var)
        data = list(zip(int_values, self.json_data))
        self.cursor.executemany("""
                insert into TestJson
                values(:int_val, :json_val)
                returning IntCol, JsonCol into :int_var, :json_var""", data)
        self.assertEqual(out_int_var.values, [[v] for v in int_values])
        self.assertEqual(out_json_var.values, [[v] for v in self.json_data])

    def test_3504_boolean(self):
        "3504 - test binding boolean values as scalar JSON values"
        data = [
            True,
            False,
            True,
            True,
            False,
            True
        ]
        self.__bind_scalar_as_json(data)

    def test_3505_strings_and_bytes(self):
        "3505 - test binding strings/bytes values as scalar JSON values"
        data = [
            "String 1",
            b"A raw value",
            "A much longer string",
            b"A much longer RAW value",
            "Short string",
            b"Y"
        ]
        self.__bind_scalar_as_json(data)

    def test_3506_datetime(self):
        "3506 - test binding dates/intervals as scalar JSON values"
        data = [
            datetime.datetime.today(),
            datetime.datetime(2004, 2, 1, 3, 4, 5),
            datetime.datetime(2020, 12, 2, 13, 29, 14),
            datetime.timedelta(8.5),
            datetime.datetime(2002, 12, 13, 9, 36, 0),
            oracledb.Timestamp(2002, 12, 13, 9, 36, 0),
            datetime.datetime(2002, 12, 13)
        ]
        self.__bind_scalar_as_json(data)

    def test_3507_bind_number(self):
        "3507 - test binding number in json values"
        data = [
            0,
            1,
            25.25,
            6088343244,
            -9999999999999999999,
            decimal.Decimal("0.25"),
            decimal.Decimal("10.25"),
            decimal.Decimal("319438950232418390.273596")
        ]
        self.__bind_scalar_as_json(data)

    def test_3508_unsupported_python_type_bind(self):
        "3508 - test binding unsupported python type with JSON"
        self.cursor.execute("truncate table TestJson")
        self.cursor.setinputsizes(None, oracledb.DB_TYPE_JSON)
        insert_sql = "insert into TestJson values (:1, :2)"
        self.assertRaisesRegex(oracledb.NotSupportedError, "^DPY-3003:",
                               self.cursor.execute, insert_sql, [1, list])

    def test_3509_unsupported_python_type_fetch(self):
        "3509 - test fetching an unsupported python type with JSON"
        sql = """
                select json(json_scalar(to_yminterval('8-04')))
                from dual"""
        self.cursor.prefetchrows = 0
        self.cursor.execute(sql)
        self.assertRaisesRegex(oracledb.NotSupportedError, "^DPY-3007:",
                               self.cursor.fetchone)

    def test_3510_fetch_all_supported_types(self):
        "3510 - fetch all supported types"
        sql = """
            select json('{
                "binary_float": {"$numberFloat": 38.75},
                "binary_double": {"$numberDouble": 125.875},
                "date_no_time": {"$oracleDate": "2022-12-05"},
                "date_with_time": {"$oracleDate": "2022-12-05T15:06:05"},
                "empty_string": "",
                "explicit_long": {"$numberLong": 9223372036854775807},
                "false": false,
                "interval_ds": {"$intervalDaySecond" : "P133DT2H5M8.123S"},
                "long_integer": 12345678901234567890123456789012345,
                "null": null,
                "short_decimal": {"$numberDecimal": 18.25},
                "short_integer": {"$numberInt": 5 },
                "short_raw": {"$rawhex": "73686f72745f726177"},
                "short_string": "A short string",
                "small_integer": 1234,
                "small_float": 25.25,
                "string_uint8": "A longer string but still < 256 bytes",
                "true": true,
                "ts_no_fs": {"$oracleTimestamp": "2022-12-06T18:12:35"},
                "ts_tz": {"$oracleTimestampTZ": "2022-12-07T22:59:15.1234Z"},
                "ts_with_fs": {"$oracleTimestamp": "2022-12-06T18:12:35.123"}
            }'
            extended) from dual"""
        expected_data = dict(
            binary_float=38.75,
            binary_double=125.875,
            date_no_time=datetime.datetime(2022, 12, 5),
            date_with_time=datetime.datetime(2022, 12, 5, 15, 6, 5),
            empty_string="",
            explicit_long=9223372036854775807,
            false=False,
            interval_ds=datetime.timedelta(days=133, seconds=7508,
                                           microseconds=123000),
            null=None,
            long_integer=12345678901234567890123456789012345,
            short_decimal=18.25,
            short_integer=5,
            short_raw=b"short_raw",
            short_string="A short string",
            small_integer=1234,
            small_float=25.25,
            string_uint8="A longer string but still < 256 bytes",
            true=True,
            ts_no_fs=datetime.datetime(2022, 12, 6, 18, 12, 35),
            ts_tz=datetime.datetime(2022, 12, 7, 22, 59, 15, 123400),
            ts_with_fs=datetime.datetime(2022, 12, 6, 18, 12, 35, 123000)
        )
        self.cursor.execute(sql)
        actual_data, = self.cursor.fetchone()
        self.assertEqual(actual_data, expected_data)

if __name__ == "__main__":
    test_env.run_test_cases()



#------------------------------------------------------------------------------
# Copyright (c) 2020, 2023, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# __init__.py
#
# Package initialization module.
#------------------------------------------------------------------------------

import sys

from .version import (
    __version__ as __version__
)

from .constants import *
from .exceptions import *

from .errors import (
    _Error as _Error
)

from .defaults import (
    defaults as defaults
)

from .connection import (
    connect as connect,
    Connection as Connection
)

from .cursor import (
    Cursor as Cursor
)

from .pool import (
    create_pool as create_pool,
    ConnectionPool as ConnectionPool
)

from .connect_params import (
    ConnectParams as ConnectParams
)

from .pool_params import (
    PoolParams as PoolParams
)

from .lob import (
    LOB as LOB
)

from .dbobject import (
    DbObject as DbObject,
    DbObjectType as DbObjectType
)

from .var import (
    Var as Var
)

from .dsn import (
    makedsn as makedsn
)

from .driver_mode import (
    is_thin_mode as is_thin_mode
)

from .base_impl import (

    # database types
    DB_TYPE_BFILE as DB_TYPE_BFILE,
    DB_TYPE_BINARY_DOUBLE as DB_TYPE_BINARY_DOUBLE,
    DB_TYPE_BINARY_FLOAT as DB_TYPE_BINARY_FLOAT,
    DB_TYPE_BINARY_INTEGER as DB_TYPE_BINARY_INTEGER,
    DB_TYPE_BLOB as DB_TYPE_BLOB,
    DB_TYPE_BOOLEAN as DB_TYPE_BOOLEAN,
    DB_TYPE_CHAR as DB_TYPE_CHAR,
    DB_TYPE_CLOB as DB_TYPE_CLOB,
    DB_TYPE_CURSOR as DB_TYPE_CURSOR,
    DB_TYPE_DATE as DB_TYPE_DATE,
    DB_TYPE_INTERVAL_DS as DB_TYPE_INTERVAL_DS,
    DB_TYPE_INTERVAL_YM as DB_TYPE_INTERVAL_YM,
    DB_TYPE_JSON as DB_TYPE_JSON,
    DB_TYPE_LONG as DB_TYPE_LONG,
    DB_TYPE_LONG_NVARCHAR as DB_TYPE_LONG_NVARCHAR,
    DB_TYPE_LONG_RAW as DB_TYPE_LONG_RAW,
    DB_TYPE_NCHAR as DB_TYPE_NCHAR,
    DB_TYPE_NCLOB as DB_TYPE_NCLOB,
    DB_TYPE_NUMBER as DB_TYPE_NUMBER,
    DB_TYPE_NVARCHAR as DB_TYPE_NVARCHAR,
    DB_TYPE_OBJECT as DB_TYPE_OBJECT,
    DB_TYPE_RAW as DB_TYPE_RAW,
    DB_TYPE_ROWID as DB_TYPE_ROWID,
    DB_TYPE_TIMESTAMP as DB_TYPE_TIMESTAMP,
    DB_TYPE_TIMESTAMP_LTZ as DB_TYPE_TIMESTAMP_LTZ,
    DB_TYPE_TIMESTAMP_TZ as DB_TYPE_TIMESTAMP_TZ,
    DB_TYPE_UNKNOWN as DB_TYPE_UNKNOWN,
    DB_TYPE_UROWID as DB_TYPE_UROWID,
    DB_TYPE_VARCHAR as DB_TYPE_VARCHAR,

    # API types
    BINARY as BINARY,
    DATETIME as DATETIME,
    NUMBER as NUMBER,
    ROWID as ROWID,
    STRING as STRING
)

from .thick_impl import (
    clientversion as clientversion,
    init_oracle_client as init_oracle_client
)

from .constructors import (
    Binary as Binary,
    Date as Date,
    DateFromTicks as DateFromTicks,
    Time as Time,
    TimeFromTicks as TimeFromTicks,
    Timestamp as Timestamp,
    TimestampFromTicks as TimestampFromTicks
)

package = sys.modules[__name__]
base_impl.init_base_impl(package)
thick_impl.init_thick_impl(package)
thin_impl.init_thin_impl(package)
del package

# future object used for managing backwards incompatible changes
class Future:

    def __getattr__(self, name):
        return None

    def __setattr__(self, name, value):
        pass

__future__ = Future()

# remove unnecessary symbols
del exceptions, errors, connection, pool, constants, driver_mode, sys
del constructors, dsn, lob, base_impl, thick_impl, thin_impl, utils, var
del connect_params, pool_params, subscr, aq, soda, cursor, dbobject

# general aliases (for backwards compatibility)
ObjectType = DbObjectType
Object = DbObject
SessionPool = ConnectionPool
version = __version__

# aliases for database types (for backwards compatibility)
BFILE = DB_TYPE_BFILE
BLOB = DB_TYPE_BLOB
BOOLEAN = DB_TYPE_BOOLEAN
CLOB = DB_TYPE_CLOB
CURSOR = DB_TYPE_CURSOR
FIXED_CHAR = DB_TYPE_CHAR
FIXED_NCHAR = DB_TYPE_NCHAR
INTERVAL = DB_TYPE_INTERVAL_DS
LONG_BINARY = DB_TYPE_LONG_RAW
LONG_STRING = DB_TYPE_LONG
NATIVE_INT = DB_TYPE_BINARY_INTEGER
NATIVE_FLOAT = DB_TYPE_BINARY_DOUBLE
NCHAR = DB_TYPE_NVARCHAR
OBJECT = DB_TYPE_OBJECT
NCLOB = DB_TYPE_NCLOB
TIMESTAMP = DB_TYPE_TIMESTAMP

# aliases for authhentication modes (for backwards compatibility)
DEFAULT_AUTH = AUTH_MODE_DEFAULT
SYSASM = AUTH_MODE_SYSASM
SYSBKP = AUTH_MODE_SYSBKP
SYSDBA = AUTH_MODE_SYSDBA
SYSDGD = AUTH_MODE_SYSDGD
SYSKMT = AUTH_MODE_SYSKMT
SYSOPER = AUTH_MODE_SYSOPER
SYSRAC = AUTH_MODE_SYSRAC
PRELIM_AUTH = AUTH_MODE_PRELIM

# aliases for pool "get" modes (for backwards compatibility)
SPOOL_ATTRVAL_WAIT = POOL_GETMODE_WAIT
SPOOL_ATTRVAL_NOWAIT = POOL_GETMODE_NOWAIT
SPOOL_ATTRVAL_FORCEGET = POOL_GETMODE_FORCEGET
SPOOL_ATTRVAL_TIMEDWAIT = POOL_GETMODE_TIMEDWAIT

# aliases for purity (for backwards compatibility)
ATTR_PURITY_DEFAULT = PURITY_DEFAULT
ATTR_PURITY_NEW = PURITY_NEW
ATTR_PURITY_SELF = PURITY_SELF

# aliases for subscription protocols (for backwards compatibility)
SUBSCR_PROTO_OCI = SUBSCR_PROTO_CALLBACK



#------------------------------------------------------------------------------
# Copyright (c) 2020, 2023, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# statement.pyx
#
# Cython file defining the Statement and BindInfo classes used to hold
# information about statements that are executed and any bind parameters that
# may be bound to those statements (embedded in thin_impl.pyx).
#------------------------------------------------------------------------------

# Rules for named binds:
# 1. Quoted and non-quoted bind names are allowed.
# 2. Quoted binds can contain any characters.
# 3. Non-quoted binds must begin with an alphabet character.
# 4. Non-quoted binds can only contain alphanumeric characters, the underscore,
#    the dollar sign and the pound sign.
# 5. Non-quoted binds cannot be Oracle Database Reserved Names (Server handles
#    this case and returns an appropriate error)
BIND_PATTERN = r':\s*((?:".*?")|(?:[^\W\d_][\w\$#]*)|\d+)'

# pattern used for detecting a DML returning clause; bind variables in the
# SQL prior to the INTO keyword are input variables; bind varibles in the SQL
# after the INTO keyword are output variables
DML_RETURNING_PATTERN = r'(?si)(?<=\bRETURNING\b)(.*?)(?=\bINTO\b)'

# patterns for identifying comments and quoted strings
SINGLE_LINE_COMMENT_PATTERN = r'--.*'
MULTI_LINE_COMMENT_PATTERN = r'(?s)/\*.*?\*/'
CONSTANT_STRING_PATTERN = r"(?s)'.*?'"
QUOTED_NAME_PATTERN = r'(:\s*)?(".*?")'

cdef class BindInfo:

    cdef:
        uint32_t num_elements
        bint _is_return_bind
        uint8_t ora_type_num
        uint32_t buffer_size
        int16_t precision
        uint8_t bind_dir
        uint32_t size
        str _bind_name
        bint is_array
        int16_t scale
        uint8_t csfrm
        ThinVarImpl _bind_var_impl

    def __cinit__(self, str name, bint is_return_bind):
        self._bind_name = name
        self._is_return_bind = is_return_bind

    cdef BindInfo copy(self):
        return BindInfo(self._bind_name, self._is_return_bind)


cdef class Statement:

    cdef:
        str _sql
        bytes _sql_bytes
        uint32_t _sql_length
        uint16_t _cursor_id
        bint _is_query
        bint _is_plsql
        bint _is_dml
        bint _is_ddl
        bint _is_returning
        list _bind_info_list
        list _fetch_vars
        list _fetch_var_impls
        object _bind_info_dict
        object _last_output_type_handler
        uint32_t _num_columns
        bint _requires_full_execute
        bint _always_full_execute
        bint _requires_define
        bint _return_to_cache
        bint _in_use

    cdef Statement copy(self):
        cdef:
            Statement copied_statement = Statement()
            object bind_info_dict
            BindInfo bind_info
        copied_statement._sql = self._sql
        copied_statement._sql_bytes = self._sql_bytes
        copied_statement._sql_length = self._sql_length
        copied_statement._is_query = self._is_query
        copied_statement._is_plsql = self._is_plsql
        copied_statement._is_dml = self._is_dml
        copied_statement._is_ddl = self._is_ddl
        copied_statement._is_returning = self._is_returning
        copied_statement._bind_info_list = \
                [bind_info.copy() for bind_info in self._bind_info_list]
        copied_statement._bind_info_dict = collections.OrderedDict()
        bind_info_dict = copied_statement._bind_info_dict
        for bind_info in copied_statement._bind_info_list:
            if bind_info._bind_name in bind_info_dict:
                bind_info_dict[bind_info._bind_name].append(bind_info)
            else:
                bind_info_dict[bind_info._bind_name] = [bind_info]
        copied_statement._return_to_cache = False
        return copied_statement

    cdef int _add_binds(self, str sql, bint is_return_bind) except -1:
        """
        Add bind information to the statement by examining the passed SQL for
        bind variable names.
        """
        cdef:
            BindInfo info
            str name
        for name in re.findall(BIND_PATTERN, sql):
            if name.startswith('"') and name.endswith('"'):
                name = name[1:-1]
            else:
                name = name.upper()
            if self._is_plsql and name in self._bind_info_dict:
                continue
            info = BindInfo(name, is_return_bind)
            self._bind_info_list.append(info)
            if info._bind_name in self._bind_info_dict:
                self._bind_info_dict[info._bind_name].append(info)
            else:
                self._bind_info_dict[info._bind_name] = [info]

    cdef int _adjust_requires_define(self) except -1:
        """
        The define step is only supposed to be done once for each cursor, but a
        full execute is required if a define was ever performed, so perform
        that adjustment here, if needed.
        """
        if self._requires_define:
            if self._always_full_execute:
                self._requires_define = False
            else:
                self._always_full_execute = True

    cdef _determine_statement_type(self, str sql):
        """
        Determine the type of the SQL statement by examining the first keyword
        found in the statement.
        """
        tokens = sql.strip().lstrip("(")[:10].split()
        if tokens:
            sql_keyword = tokens[0].upper()
            if sql_keyword in ("DECLARE", "BEGIN", "CALL"):
                self._is_plsql = True
            elif sql_keyword in ("SELECT", "WITH"):
                self._is_query = True
            elif sql_keyword in ("INSERT", "UPDATE", "DELETE", "MERGE"):
                self._is_dml = True
            elif sql_keyword in ("CREATE", "ALTER", "DROP", "TRUNCATE"):
                self._is_ddl = True

    cdef int _prepare(self, str sql) except -1:
        """
        Prepare the SQL for execution by determining the list of bind names
        that are found within it. The length of the SQL text is also calculated
        at this time. If the character sets of the client and server are
        identical, the length is calculated in bytes; otherwise, the length is
        calculated in characters.
        """
        cdef:
            str input_sql, returning_sql = None
            object match

        # retain normalized SQL (as string and bytes) as well as the length
        self._sql = sql
        self._sql_bytes = self._sql.encode()
        self._sql_length = <uint32_t> len(self._sql_bytes)

        # create empty list (bind by position) and dict (bind by name)
        self._bind_info_dict = collections.OrderedDict()
        self._bind_info_list = []

        # Strip single/multiline comments and replace constant strings and
        # quoted names with single characters in order to facilitate detection
        # of bind variables; note that bind variables can be quoted so a check
        # must be made to ensure that a quoted string doesn't refer to a bind
        # variable first before it can be replaced
        sql = re.sub(MULTI_LINE_COMMENT_PATTERN, "", sql)
        sql = re.sub(SINGLE_LINE_COMMENT_PATTERN, "", sql)
        sql = re.sub(CONSTANT_STRING_PATTERN, "S", sql)
        sql = re.sub(QUOTED_NAME_PATTERN,
                    lambda m: m.group(0) if sql[m.start(0)] == ":" else "Q",
                    sql)

        # determine statement type
        self._determine_statement_type(sql)

        # bind variables can only be found in queries, DML and PL/SQL
        if self._is_query or self._is_dml or self._is_plsql:
            input_sql = sql
            if self._is_dml:
                match = re.search(DML_RETURNING_PATTERN, sql)
                if match is not None:
                    pos = match.end()
                    input_sql = sql[:pos]
                    returning_sql = sql[pos + 4:]
            self._add_binds(input_sql, is_return_bind=False)
            if returning_sql is not None:
                self._is_returning = True
                self._add_binds(returning_sql, is_return_bind=True)

    cdef int _set_var(self, BindInfo bind_info, ThinVarImpl var_impl,
                      ThinCursorImpl cursor_impl) except -1:
        """
        Set the variable on the bind information and copy across metadata that
        will be used for binding. If the bind metadata has changed, mark the
        statement as requiring a full execute. In addition, binding a REF
        cursor also requires a full execute.
        """
        cdef object value
        if var_impl.dbtype._ora_type_num == TNS_DATA_TYPE_CURSOR:
            for value in var_impl._values:
                if value is not None and value._impl is cursor_impl:
                    errors._raise_err(errors.ERR_SELF_BIND_NOT_SUPPORTED)
            self._requires_full_execute = True
        if var_impl.dbtype._ora_type_num != bind_info.ora_type_num \
                or var_impl.size != bind_info.size \
                or var_impl.buffer_size != bind_info.buffer_size \
                or var_impl.precision != bind_info.precision \
                or var_impl.scale != bind_info.scale \
                or var_impl.is_array != bind_info.is_array \
                or var_impl.num_elements != bind_info.num_elements \
                or var_impl.dbtype._csfrm != bind_info.csfrm:
            bind_info.ora_type_num = var_impl.dbtype._ora_type_num
            bind_info.csfrm = var_impl.dbtype._csfrm
            bind_info.is_array = var_impl.is_array
            bind_info.num_elements = var_impl.num_elements
            bind_info.size = var_impl.size
            bind_info.buffer_size = var_impl.buffer_size
            bind_info.precision = var_impl.precision
            bind_info.scale = var_impl.scale
            self._requires_full_execute = True
        bind_info._bind_var_impl = var_impl



#------------------------------------------------------------------------------
# Copyright (c) 2020, 2023, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

"""
1700 - Module for testing error objects
"""

import pickle

import oracledb
import test_env

class TestCase(test_env.BaseTestCase):

    def test_1700_parse_error(self):
        "1700 - test parse error returns offset correctly"
        with self.assertRaises(oracledb.Error) as cm:
            self.cursor.execute("begin t_Missing := 5; end;")
        error_obj, = cm.exception.args
        self.assertEqual(error_obj.full_code, "ORA-06550")
        self.assertEqual(error_obj.offset, 6)

    def test_1701_pickle_error(self):
        "1701 - test picking/unpickling an error object"
        with self.assertRaises(oracledb.Error) as cm:
            self.cursor.execute("""
                    begin
                        raise_application_error(-20101, 'Test!');
                    end;""")
        error_obj, = cm.exception.args
        self.assertEqual(type(error_obj), oracledb._Error)
        self.assertTrue("Test!" in error_obj.message)
        self.assertEqual(error_obj.code, 20101)
        self.assertEqual(error_obj.offset, 0)
        self.assertTrue(isinstance(error_obj.isrecoverable, bool))
        new_error_obj = pickle.loads(pickle.dumps(error_obj))
        self.assertEqual(type(new_error_obj), oracledb._Error)
        self.assertTrue(new_error_obj.message == error_obj.message)
        self.assertTrue(new_error_obj.code == error_obj.code)
        self.assertTrue(new_error_obj.offset == error_obj.offset)
        self.assertTrue(new_error_obj.context == error_obj.context)
        self.assertTrue(new_error_obj.isrecoverable == error_obj.isrecoverable)

    def test_1702_error_full_code(self):
        "1702 - test generation of full_code for ORA, DPI and DPY errors"
        cursor = self.connection.cursor()
        with self.assertRaises(oracledb.Error) as cm:
            cursor.execute(None)
        error_obj, = cm.exception.args
        self.assertEqual(error_obj.full_code, "DPY-2001")
        if not self.connection.thin:
            with self.assertRaises(oracledb.Error) as cm:
                cursor.execute("truncate table TestTempTable")
                int_var = cursor.var(int)
                str_var = cursor.var(str, 2)
                cursor.execute("""
                        insert into TestTempTable (IntCol, StringCol1)
                        values (1, 'Longer than two chars')
                        returning IntCol, StringCol1
                        into :int_var, :str_var""",
                        int_var=int_var, str_var=str_var)
            error_obj, = cm.exception.args
            self.assertEqual(error_obj.full_code, "DPI-1037")

if __name__ == "__main__":
    test_env.run_test_cases()



#------------------------------------------------------------------------------
# Copyright (c) 2020, 2022, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# json.pyx
#
# Cython file defining methods and classes used for processing JSON data
# (embedded in thick_impl.pyx).
#------------------------------------------------------------------------------

cdef class JsonBuffer:
    cdef:
        dpiJsonNode _top_node
        dpiDataBuffer _top_node_buf
        list _buffers

    def __dealloc__(self):
        _free_node(&self._top_node)

    cdef int _add_buf(self, object value, char **ptr,
                      uint32_t *length) except -1:
        cdef StringBuffer buf = StringBuffer()
        buf.set_value(value)
        if self._buffers is None:
            self._buffers = []
        self._buffers.append(buf)
        ptr[0] = buf.ptr
        length[0] = buf.length

    cdef int _populate_array_node(self, dpiJsonNode *node,
                                  list value) except -1:
        cdef:
            dpiJsonArray *array
            object child_value
            uint32_t i
        node.oracleTypeNum = DPI_ORACLE_TYPE_JSON_ARRAY
        node.nativeTypeNum = DPI_NATIVE_TYPE_JSON_ARRAY
        array = &node.value.asJsonArray
        array.numElements = <uint32_t> len(value)
        array.elements = <dpiJsonNode*> _calloc(array.numElements,
                                                sizeof(dpiJsonNode))
        array.elementValues = <dpiDataBuffer*> _calloc(array.numElements,
                                                       sizeof(dpiDataBuffer))
        for i, child_value in enumerate(value):
            array.elements[i].value = &array.elementValues[i]
            self._populate_node(&array.elements[i], child_value)

    cdef int _populate_obj_node(self, dpiJsonNode *node, dict value) except -1:
        cdef:
            object child_key, child_value
            dpiJsonObject *obj
            uint32_t i
        node.oracleTypeNum = DPI_ORACLE_TYPE_JSON_OBJECT
        node.nativeTypeNum = DPI_NATIVE_TYPE_JSON_OBJECT
        obj = &node.value.asJsonObject
        obj.numFields = <uint32_t> len(value)
        obj.fieldNames = <char**> _calloc(obj.numFields, sizeof(char*))
        obj.fieldNameLengths = <uint32_t*> _calloc(obj.numFields,
                                                   sizeof(uint32_t))
        obj.fields = <dpiJsonNode*> _calloc(obj.numFields, sizeof(dpiJsonNode))
        obj.fieldValues = <dpiDataBuffer*> _calloc(obj.numFields,
                                                   sizeof(dpiDataBuffer))
        i = 0
        for child_key, child_value in value.items():
            obj.fields[i].value = &obj.fieldValues[i]
            self._add_buf(child_key, &obj.fieldNames[i],
                          &obj.fieldNameLengths[i])
            self._populate_node(&obj.fields[i], child_value)
            i += 1

    cdef int _populate_node(self, dpiJsonNode *node, object value) except -1:
        cdef:
            dpiTimestamp *timestamp
            dpiIntervalDS *interval
            int seconds
        if value is None:
            node.oracleTypeNum = DPI_ORACLE_TYPE_NONE
            node.nativeTypeNum = DPI_NATIVE_TYPE_NULL
        elif isinstance(value, list):
            self._populate_array_node(node, value)
        elif isinstance(value, dict):
            self._populate_obj_node(node, value)
        elif isinstance(value, str):
            node.oracleTypeNum = DPI_ORACLE_TYPE_VARCHAR
            node.nativeTypeNum = DPI_NATIVE_TYPE_BYTES
            self._add_buf(value, &node.value.asBytes.ptr,
                          &node.value.asBytes.length)
        elif isinstance(value, bytes):
            node.oracleTypeNum = DPI_ORACLE_TYPE_RAW
            node.nativeTypeNum = DPI_NATIVE_TYPE_BYTES
            self._add_buf(value, &node.value.asBytes.ptr,
                          &node.value.asBytes.length)
        elif isinstance(value, bool):
            node.oracleTypeNum = DPI_ORACLE_TYPE_BOOLEAN
            node.nativeTypeNum = DPI_NATIVE_TYPE_BOOLEAN
            node.value.asBoolean = <bint> value
        elif isinstance(value, (int, PY_TYPE_DECIMAL)):
            node.oracleTypeNum = DPI_ORACLE_TYPE_NUMBER
            node.nativeTypeNum = DPI_NATIVE_TYPE_BYTES
            self._add_buf(str(value), &node.value.asBytes.ptr,
                          &node.value.asBytes.length)
        elif isinstance(value, float):
            node.oracleTypeNum = DPI_ORACLE_TYPE_NUMBER
            node.nativeTypeNum = DPI_NATIVE_TYPE_DOUBLE
            node.value.asDouble = <double> value
        elif isinstance(value, PY_TYPE_DATETIME):
            node.oracleTypeNum = DPI_ORACLE_TYPE_TIMESTAMP
            node.nativeTypeNum = DPI_NATIVE_TYPE_TIMESTAMP
            memset(&node.value.asTimestamp, 0, sizeof(node.value.asTimestamp))
            timestamp = &node.value.asTimestamp
            timestamp.year = cydatetime.datetime_year(value)
            timestamp.month = cydatetime.datetime_month(value)
            timestamp.day = cydatetime.datetime_day(value)
            timestamp.hour = cydatetime.datetime_hour(value)
            timestamp.minute = cydatetime.datetime_minute(value)
            timestamp.second = cydatetime.datetime_second(value)
            timestamp.fsecond = cydatetime.datetime_microsecond(value) * 1000
        elif isinstance(value, PY_TYPE_DATE):
            node.oracleTypeNum = DPI_ORACLE_TYPE_DATE
            node.nativeTypeNum = DPI_NATIVE_TYPE_TIMESTAMP
            memset(&node.value.asTimestamp, 0, sizeof(node.value.asTimestamp))
            timestamp = &node.value.asTimestamp
            timestamp = &node.value.asTimestamp
            timestamp.year = cydatetime.date_year(value)
            timestamp.month = cydatetime.date_month(value)
            timestamp.day = cydatetime.date_day(value)
        elif isinstance(value, datetime.timedelta):
            node.oracleTypeNum = DPI_ORACLE_TYPE_INTERVAL_DS
            node.nativeTypeNum = DPI_NATIVE_TYPE_INTERVAL_DS
            interval = &node.value.asIntervalDS
            seconds = cydatetime.timedelta_seconds(value)
            interval.days = cydatetime.timedelta_days(value)
            interval.hours = seconds // 3600
            seconds = seconds % 3600
            interval.minutes = seconds // 60
            interval.seconds = seconds % 60
            interval.fseconds = cydatetime.timedelta_microseconds(value) * 1000
        else:
            errors._raise_err(errors.ERR_PYTHON_TYPE_NOT_SUPPORTED,
                              typ=type(value))

    cdef int from_object(self, object value) except -1:
        self._top_node.value = &self._top_node_buf
        self._populate_node(&self._top_node, value)


cdef void* _calloc(size_t num_elements, size_t size) except NULL:
    cdef:
        size_t num_bytes = num_elements * size
        void *ptr
    ptr = cpython.PyMem_Malloc(num_bytes)
    memset(ptr, 0, num_bytes)
    return ptr

cdef void _free_node(dpiJsonNode *node):
    cdef:
        dpiJsonArray *array
        dpiJsonObject *obj
        uint32_t i
    if node.nativeTypeNum == DPI_NATIVE_TYPE_JSON_ARRAY:
        array = &node.value.asJsonArray
        if array.elements != NULL:
            for i in range(array.numElements):
                if array.elements[i].value != NULL:
                    _free_node(&array.elements[i])
            cpython.PyMem_Free(array.elements)
            array.elements = NULL
        if array.elementValues != NULL:
            cpython.PyMem_Free(array.elementValues)
            array.elementValues = NULL
    elif node.nativeTypeNum == DPI_NATIVE_TYPE_JSON_OBJECT:
        obj = &node.value.asJsonObject
        if obj.fields != NULL:
            for i in range(obj.numFields):
                if obj.fields[i].value != NULL:
                    _free_node(&obj.fields[i])
            cpython.PyMem_Free(obj.fields)
            obj.fields = NULL
        if obj.fieldNames != NULL:
            cpython.PyMem_Free(obj.fieldNames)
            obj.fieldNames = NULL
        if obj.fieldNameLengths != NULL:
            cpython.PyMem_Free(obj.fieldNameLengths)
            obj.fieldNameLengths = NULL
        if obj.fieldValues != NULL:
            cpython.PyMem_Free(obj.fieldValues)
            obj.fieldValues = NULL



#------------------------------------------------------------------------------
# Copyright (c) 2020, 2022, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# lob.pyx
#
# Cython file defining the base Lob implementation class (embedded in
# base_impl.pyx).
#------------------------------------------------------------------------------

cdef class BaseLobImpl:

    @utils.CheckImpls("closing a LOB")
    def close(self):
        pass

    @utils.CheckImpls("checking if a BFILE exists")
    def file_exists(self):
        pass

    @utils.CheckImpls("freeing the lob object")
    def free_lob(self):
        pass

    @utils.CheckImpls("getting the chunk size of a LOB")
    def get_chunk_size(self):
        pass

    @utils.CheckImpls("getting the file name and directory alias of a BFILE")
    def get_file_name(self):
        pass

    @utils.CheckImpls("getting whether a LOB is open or not")
    def get_is_open(self):
        pass

    @utils.CheckImpls("getting the maximum amount that can be read from a LOB")
    def get_max_amount(self):
        pass

    @utils.CheckImpls("getting the size of a LOB")
    def get_size(self):
        pass

    @utils.CheckImpls("opening a LOB")
    def open(self):
        pass

    @utils.CheckImpls("reading data from a LOB")
    def read(self, uint64_t offset, uint64_t amount):
        pass

    @utils.CheckImpls("setting the file name an directory alias of a BFILE")
    def set_file_name(self, str dir_alias, str name):
        pass

    @utils.CheckImpls("trimming a LOB")
    def trim(self, uint64_t new_size):
        pass

    @utils.CheckImpls("writing data to a LOB")
    def write(self, object value, uint64_t offset):
        pass



#------------------------------------------------------------------------------
# Copyright (c) 2020, 2022, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

"""
1400 - Module for testing date/time variables
"""

import datetime
import time

import oracledb
import test_env

class TestCase(test_env.BaseTestCase):

    def setUp(self):
        super().setUp()
        self.raw_data = []
        self.data_by_key = {}
        for i in range(1, 11):
            time_tuple = (2002, 12, 9, 0, 0, 0, 0, 0, -1)
            time_in_ticks = time.mktime(time_tuple) + i * 86400 + i * 8640
            date_col = oracledb.TimestampFromTicks(int(time_in_ticks))
            if i % 2:
                time_in_ticks = time.mktime(time_tuple) + i * 86400 * 2 + \
                        i * 12960
                nullable_col = oracledb.TimestampFromTicks(int(time_in_ticks))
            else:
                nullable_col = None
            tuple = (i, date_col, nullable_col)
            self.raw_data.append(tuple)
            self.data_by_key[i] = tuple

    def test_1400_bind_date(self):
        "1400 - test binding in a date"
        self.cursor.execute("""
                select * from TestDates
                where DateCol = :value""",
                value = oracledb.Timestamp(2002, 12, 13, 9, 36, 0))
        self.assertEqual(self.cursor.fetchall(), [self.data_by_key[4]])

    def test_1401_bind_datetime(self):
        "1401 - test binding in a datetime.datetime value"
        self.cursor.execute("""
                select * from TestDates
                where DateCol = :value""",
                value=datetime.datetime(2002, 12, 13, 9, 36, 0))
        self.assertEqual(self.cursor.fetchall(), [self.data_by_key[4]])

    def test_1402_bind_date_in_datetime_var(self):
        "1402 - test binding date in a datetime variable"
        var = self.cursor.var(oracledb.DATETIME)
        date_val = datetime.date.today()
        var.setvalue(0, date_val)
        self.cursor.execute("select :1 from dual", [var])
        result, = self.cursor.fetchone()
        self.assertEqual(result.date(), date_val)

    def test_1403_bind_date_after_string(self):
        "1403 - test binding in a date after setting input sizes to a string"
        self.cursor.setinputsizes(value=15)
        self.cursor.execute("""
                select * from TestDates
                where DateCol = :value""",
                value = oracledb.Timestamp(2002, 12, 14, 12, 0, 0))
        self.assertEqual(self.cursor.fetchall(), [self.data_by_key[5]])

    def test_1404_bind_null(self):
        "1404 - test binding in a null"
        self.cursor.setinputsizes(value=oracledb.DATETIME)
        self.cursor.execute("""
                select * from TestDates
                where DateCol = :value""",
                value = None)
        self.assertEqual(self.cursor.fetchall(), [])

    def test_1405_bind_date_array_direct(self):
        "1405 - test binding in a date array"
        return_value = self.cursor.var(oracledb.NUMBER)
        array = [r[1] for r in self.raw_data]
        statement = """
                begin
                    :return_value := pkg_TestDateArrays.TestInArrays(
                        :start_value, :base_date, :array);
                end;"""
        self.cursor.execute(statement, return_value=return_value,
                            start_value=5,
                            base_date=oracledb.Date(2002, 12, 12), array=array)
        self.assertEqual(return_value.getvalue(), 35.5)
        array = array + array[:5]
        self.cursor.execute(statement, start_value=7,
                            base_date=oracledb.Date(2002, 12, 13), array=array)
        self.assertEqual(return_value.getvalue(), 24.0)

    def test_1406_bind_date_array_by_sizes(self):
        "1406 - test binding in a date array (with setinputsizes)"
        return_value = self.cursor.var(oracledb.NUMBER)
        self.cursor.setinputsizes(array=[oracledb.DATETIME, 10])
        array = [r[1] for r in self.raw_data]
        self.cursor.execute("""
                begin
                    :return_value := pkg_TestDateArrays.TestInArrays(
                        :start_value, :base_date, :array);
                end;""",
                return_value=return_value,
                start_value=6,
                base_date=oracledb.Date(2002, 12, 13),
                array=array)
        self.assertEqual(return_value.getvalue(), 26.5)

    def test_1407_bind_date_array_by_var(self):
        "1407 - test binding in a date array (with arrayvar)"
        return_value = self.cursor.var(oracledb.NUMBER)
        array = self.cursor.arrayvar(oracledb.DATETIME, 10, 20)
        array.setvalue(0, [r[1] for r in self.raw_data])
        self.cursor.execute("""
                begin
                    :return_value := pkg_TestDateArrays.TestInArrays(
                        :start_value, :base_date, :array);
                end;""",
                return_value=return_value,
                start_value=7,
                base_date=oracledb.Date(2002, 12, 14),
                array=array)
        self.assertEqual(return_value.getvalue(), 17.5)

    def test_1408_bind_in_out_date_array_by_var(self):
        "1408 - test binding in/out a date array (with arrayvar)"
        array = self.cursor.arrayvar(oracledb.DATETIME, 10, 100)
        original_data = [r[1] for r in self.raw_data]
        array.setvalue(0, original_data)
        self.cursor.execute("""
                begin
                    pkg_TestDateArrays.TestInOutArrays(:num_elems, :array);
                end;""",
                num_elems=5,
                array=array)
        self.assertEqual(array.getvalue(),
                [ oracledb.Timestamp(2002, 12, 17, 2, 24, 0),
                  oracledb.Timestamp(2002, 12, 18, 4, 48, 0),
                  oracledb.Timestamp(2002, 12, 19, 7, 12, 0),
                  oracledb.Timestamp(2002, 12, 20, 9, 36, 0),
                  oracledb.Timestamp(2002, 12, 21, 12, 0, 0) ] + \
                original_data[5:])

    def test_1409_bind_out_date_array_by_var(self):
        "1409 - test binding out a date array (with arrayvar)"
        array = self.cursor.arrayvar(oracledb.DATETIME, 6, 100)
        self.cursor.execute("""
                begin
                    pkg_TestDateArrays.TestOutArrays(:num_elems, :array);
                end;""",
                num_elems=6,
                array=array)
        self.assertEqual(array.getvalue(),
                [ oracledb.Timestamp(2002, 12, 13, 4, 48, 0),
                  oracledb.Timestamp(2002, 12, 14, 9, 36, 0),
                  oracledb.Timestamp(2002, 12, 15, 14, 24, 0),
                  oracledb.Timestamp(2002, 12, 16, 19, 12, 0),
                  oracledb.Timestamp(2002, 12, 18, 0, 0, 0),
                  oracledb.Timestamp(2002, 12, 19, 4, 48, 0) ])

    def test_1410_bind_out_set_input_sizes(self):
        "1410 - test binding out with set input sizes defined"
        bind_vars = self.cursor.setinputsizes(value=oracledb.DATETIME)
        self.cursor.execute("""
                begin
                    :value := to_date(20021209, 'YYYYMMDD');
                end;""")
        self.assertEqual(bind_vars["value"].getvalue(),
                         oracledb.Timestamp(2002, 12, 9))

    def test_1411_bind_in_out_set_input_sizes(self):
        "1411 - test binding in/out with set input sizes defined"
        bind_vars = self.cursor.setinputsizes(value=oracledb.DATETIME)
        self.cursor.execute("""
                begin
                    :value := :value + 5.25;
                end;""",
                value=oracledb.Timestamp(2002, 12, 12, 10, 0, 0))
        self.assertEqual(bind_vars["value"].getvalue(),
                         oracledb.Timestamp(2002, 12, 17, 16, 0, 0))

    def test_1412_bind_out_var(self):
        "1412 - test binding out with cursor.var() method"
        var = self.cursor.var(oracledb.DATETIME)
        self.cursor.execute("""
                begin
                    :value := to_date('20021231 12:31:00',
                        'YYYYMMDD HH24:MI:SS');
                end;""",
                value=var)
        self.assertEqual(var.getvalue(),
                         oracledb.Timestamp(2002, 12, 31, 12, 31, 0))

    def test_1413_bind_in_out_var_direct_set(self):
        "1413 - test binding in/out with cursor.var() method"
        var = self.cursor.var(oracledb.DATETIME)
        var.setvalue(0, oracledb.Timestamp(2002, 12, 9, 6, 0, 0))
        self.cursor.execute("""
                begin
                    :value := :value + 5.25;
                end;""",
                value=var)
        self.assertEqual(var.getvalue(),
                         oracledb.Timestamp(2002, 12, 14, 12, 0, 0))

    def test_1414_cursor_description(self):
        "1414 - test cursor description is accurate"
        self.cursor.execute("select * from TestDates")
        expected_value = [
            ('INTCOL', oracledb.DB_TYPE_NUMBER, 10, None, 9, 0, False),
            ('DATECOL', oracledb.DB_TYPE_DATE, 23, None, None, None, False),
            ('NULLABLECOL', oracledb.DB_TYPE_DATE, 23, None, None, None, True)
        ]
        self.assertEqual(self.cursor.description, expected_value)

    def test_1415_fetchall(self):
        "1415 - test that fetching all of the data returns the correct results"
        self.cursor.execute("select * From TestDates order by IntCol")
        self.assertEqual(self.cursor.fetchall(), self.raw_data)
        self.assertEqual(self.cursor.fetchall(), [])

    def test_1416_fetchmany(self):
        "1416 - test that fetching data in chunks returns the correct results"
        self.cursor.execute("select * From TestDates order by IntCol")
        self.assertEqual(self.cursor.fetchmany(3), self.raw_data[0:3])
        self.assertEqual(self.cursor.fetchmany(2), self.raw_data[3:5])
        self.assertEqual(self.cursor.fetchmany(4), self.raw_data[5:9])
        self.assertEqual(self.cursor.fetchmany(3), self.raw_data[9:])
        self.assertEqual(self.cursor.fetchmany(3), [])

    def test_1417_fetchone(self):
        "1417 - test that fetching a single row returns the correct results"
        self.cursor.execute("""
                select *
                from TestDates
                where IntCol in (3, 4)
                order by IntCol""")
        self.assertEqual(self.cursor.fetchone(), self.data_by_key[3])
        self.assertEqual(self.cursor.fetchone(), self.data_by_key[4])
        self.assertEqual(self.cursor.fetchone(), None)

if __name__ == "__main__":
    test_env.run_test_cases()



#------------------------------------------------------------------------------
# Copyright (c) 2021, 2023, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

"""
3700 - Module for testing all variable types.
"""

import datetime
import decimal
import unittest

import oracledb
import test_env

class TestCase(test_env.BaseTestCase):

    def _test_positive_set_and_get(self, var_type, value_to_set,
                                   expected_value, type_name=None):
        var = self.cursor.var(var_type, typename=type_name)
        var.setvalue(0, value_to_set)
        result = var.getvalue()
        if isinstance(result, oracledb.LOB):
            result = result.read()
        elif isinstance(result, oracledb.DbObject):
            result = self.get_db_object_as_plain_object(result)
        if isinstance(expected_value, datetime.date) \
                and not isinstance(expected_value, datetime.datetime):
            if isinstance(result, datetime.datetime):
                result = result.date()
        self.assertEqual(type(result), type(expected_value))
        self.assertEqual(result, expected_value)

    def _test_negative_set_and_get(self, var_type, value_to_set,
                                   type_name=None):
        var = self.cursor.var(var_type, typename=type_name)
        self.assertRaises((TypeError, oracledb.DatabaseError),
                           var.setvalue, 0, value_to_set)

    def test_3700_DB_TYPE_NUMBER(self):
        "3700 - setting values on variables of type DB_TYPE_NUMBER"
        self._test_positive_set_and_get(int, 5, 5)
        self._test_positive_set_and_get(oracledb.DB_TYPE_NUMBER, 3.5, 3.5)
        self._test_positive_set_and_get(decimal.Decimal,
                                        decimal.Decimal("24.8"),
                                        decimal.Decimal("24.8"))
        self._test_positive_set_and_get(int, True, 1)
        self._test_positive_set_and_get(int, False, 0)
        self._test_positive_set_and_get(oracledb.DB_TYPE_NUMBER, None, None)
        self._test_negative_set_and_get(oracledb.DB_TYPE_NUMBER, "abc")

    def test_3701_DB_TYPE_BINARY_INTEGER(self):
        "3701 - setting values on variables of type DB_TYPE_BINARY_INTEGER"
        self._test_positive_set_and_get(oracledb.DB_TYPE_BINARY_INTEGER, 5, 5)
        self._test_positive_set_and_get(oracledb.DB_TYPE_BINARY_INTEGER,
                                        3.5, 3)
        self._test_positive_set_and_get(oracledb.DB_TYPE_BINARY_INTEGER,
                                        decimal.Decimal("24.8"), 24)
        self._test_positive_set_and_get(oracledb.DB_TYPE_BINARY_INTEGER,
                                        True, 1)
        self._test_positive_set_and_get(oracledb.DB_TYPE_BINARY_INTEGER,
                                        False, 0)
        self._test_positive_set_and_get(oracledb.DB_TYPE_BINARY_INTEGER,
                                        None, None)
        self._test_negative_set_and_get(oracledb.DB_TYPE_BINARY_INTEGER, "abc")

    def test_3702_DB_TYPE_VARCHAR(self):
        "3702 - setting values on variables of type DB_TYPE_VARCHAR"
        value = "A VARCHAR string"
        self._test_positive_set_and_get(oracledb.DB_TYPE_VARCHAR, value, value)
        value = b"A raw string for VARCHAR"
        self._test_positive_set_and_get(oracledb.DB_TYPE_VARCHAR, value,
                                        value.decode())
        self._test_positive_set_and_get(oracledb.DB_TYPE_VARCHAR, None, None)
        self._test_negative_set_and_get(oracledb.DB_TYPE_VARCHAR, 5)

    def test_3703_DB_TYPE_NVARCHAR(self):
        "3703 - setting values on variables of type DB_TYPE_NVARCHAR"
        value = "A NVARCHAR string"
        self._test_positive_set_and_get(oracledb.DB_TYPE_NVARCHAR, value,
                                        value)
        value = b"A raw string for NVARCHAR"
        self._test_positive_set_and_get(oracledb.DB_TYPE_NVARCHAR, value,
                                        value.decode())
        self._test_positive_set_and_get(oracledb.DB_TYPE_NVARCHAR, None, None)
        self._test_negative_set_and_get(oracledb.DB_TYPE_NVARCHAR, 5)

    def test_3704_DB_TYPE_CHAR(self):
        "3704 - setting values on variables of type DB_TYPE_CHAR"
        value = "A CHAR string"
        self._test_positive_set_and_get(oracledb.DB_TYPE_CHAR, value, value)
        value = b"A raw string for CHAR"
        self._test_positive_set_and_get(oracledb.DB_TYPE_CHAR, value,
                                        value.decode())
        self._test_positive_set_and_get(oracledb.DB_TYPE_CHAR, None, None)
        self._test_negative_set_and_get(oracledb.DB_TYPE_CHAR, 5)

    def test_3705_DB_TYPE_NCHAR(self):
        "3705 - setting values on variables of type DB_TYPE_NCHAR"
        value = "A NCHAR string"
        self._test_positive_set_and_get(oracledb.DB_TYPE_NCHAR, value, value)
        value = b"A raw string for NCHAR"
        self._test_positive_set_and_get(oracledb.DB_TYPE_CHAR, value,
                                        value.decode())
        self._test_positive_set_and_get(oracledb.DB_TYPE_NCHAR, None, None)
        self._test_negative_set_and_get(oracledb.DB_TYPE_NCHAR, 5)

    def test_3706_DB_TYPE_LONG(self):
        "3706 - setting values on variables of type DB_TYPE_LONG"
        value = "Long Data" * 15000
        self._test_positive_set_and_get(oracledb.DB_TYPE_LONG, value, value)
        value = b"Raw data for LONG" * 15000
        self._test_positive_set_and_get(oracledb.DB_TYPE_LONG, value,
                                        value.decode())
        self._test_positive_set_and_get(oracledb.DB_TYPE_LONG, None, None)
        self._test_negative_set_and_get(oracledb.DB_TYPE_LONG, 5)

    def test_3707_DB_TYPE_RAW(self):
        "3707 - setting values on variables of type DB_TYPE_RAW"
        value = b'Raw Data'
        self._test_positive_set_and_get(oracledb.DB_TYPE_RAW, value, value)
        value = "String data for RAW"
        self._test_positive_set_and_get(oracledb.DB_TYPE_RAW, value,
                                        value.encode())
        self._test_positive_set_and_get(oracledb.DB_TYPE_RAW, None, None)
        self._test_negative_set_and_get(oracledb.DB_TYPE_RAW, 5)

    def test_3708_DB_TYPE_LONG_RAW(self):
        "3708 - setting values on variables of type DB_TYPE_LONG_RAW"
        value = b'Long Raw Data' * 15000
        self._test_positive_set_and_get(oracledb.DB_TYPE_LONG_RAW, value,
                                        value)
        value = "String data for LONG RAW" * 15000
        self._test_positive_set_and_get(oracledb.DB_TYPE_LONG_RAW, value,
                                        value.encode())
        self._test_positive_set_and_get(oracledb.DB_TYPE_LONG_RAW, None, None)
        self._test_negative_set_and_get(oracledb.DB_TYPE_LONG_RAW, 5)

    def test_3709_DB_TYPE_DATE(self):
        "3709 - setting values on variables of type DB_TYPE_DATE"
        value = datetime.date(2017, 5, 6)
        self._test_positive_set_and_get(oracledb.DB_TYPE_DATE, value,
                                        value)
        value = datetime.datetime(2017, 5, 6, 9, 36, 0)
        self._test_positive_set_and_get(oracledb.DB_TYPE_DATE, value, value)
        self._test_positive_set_and_get(oracledb.DB_TYPE_DATE, None, None)
        self._test_negative_set_and_get(oracledb.DB_TYPE_DATE, 5)

    def test_3710_DB_TYPE_TIMESTAMP(self):
        "3710 - setting values on variables of type DB_TYPE_TIMESTAMP"
        value = datetime.date(2017, 5, 6)
        self._test_positive_set_and_get(oracledb.DB_TYPE_TIMESTAMP, value,
                                        value)
        value = datetime.datetime(2017, 5, 6, 9, 36, 0, 300000)
        self._test_positive_set_and_get(oracledb.DB_TYPE_TIMESTAMP, value,
                                        value)
        self._test_positive_set_and_get(oracledb.DB_TYPE_TIMESTAMP, None, None)
        self._test_negative_set_and_get(oracledb.DB_TYPE_TIMESTAMP, 5)

    def test_3711_DB_TYPE_TIMESTAMP_TZ(self):
        "3711 - setting values on variables of type DB_TYPE_TIMESTAMP_TZ"
        value = datetime.date(2017, 5, 6)
        self._test_positive_set_and_get(oracledb.DB_TYPE_TIMESTAMP_TZ, value,
                                        value)
        value = datetime.datetime(2017, 5, 6, 9, 36, 0, 300000)
        self._test_positive_set_and_get(oracledb.DB_TYPE_TIMESTAMP_TZ, value,
                                        value)
        self._test_positive_set_and_get(oracledb.DB_TYPE_TIMESTAMP_TZ,
                                        None, None)
        self._test_negative_set_and_get(oracledb.DB_TYPE_TIMESTAMP_TZ, 5)

    def test_3712_DB_TYPE_TIMESTAMP_LTZ(self):
        "3712 - setting values on variables of type DB_TYPE_TIMESTAMP_LTZ"
        value = datetime.date(2017, 5, 6)
        self._test_positive_set_and_get(oracledb.DB_TYPE_TIMESTAMP_LTZ, value,
                                        value)
        value = datetime.datetime(2017, 5, 6, 9, 36, 0, 300000)
        self._test_positive_set_and_get(oracledb.DB_TYPE_TIMESTAMP_LTZ, value,
                                        value)
        self._test_positive_set_and_get(oracledb.DB_TYPE_TIMESTAMP_LTZ,
                                        None, None)
        self._test_negative_set_and_get(oracledb.DB_TYPE_TIMESTAMP_LTZ, 5)

    def test_3713_DB_TYPE_BLOB(self):
        "3713 - setting values on variables of type DB_TYPE_BLOB"
        value = b'Short temp BLOB value'
        temp_blob = self.connection.createlob(oracledb.DB_TYPE_BLOB)
        temp_blob.write(value)
        self._test_positive_set_and_get(oracledb.DB_TYPE_BLOB, temp_blob,
                                        value)
        self._test_negative_set_and_get(oracledb.DB_TYPE_CLOB, temp_blob)
        self._test_negative_set_and_get(oracledb.DB_TYPE_NCLOB, temp_blob)
        value = b'Short BLOB value'
        self._test_positive_set_and_get(oracledb.DB_TYPE_BLOB, value, value)
        self._test_positive_set_and_get(oracledb.DB_TYPE_BLOB, None, None)
        self._test_negative_set_and_get(oracledb.DB_TYPE_BLOB, 5)

    def test_3714_DB_TYPE_CLOB(self):
        "3714 - setting values on variables of type DB_TYPE_CLOB"
        value = 'Short temp CLOB value'
        temp_clob = self.connection.createlob(oracledb.DB_TYPE_CLOB)
        temp_clob.write(value)
        self._test_positive_set_and_get(oracledb.DB_TYPE_CLOB, temp_clob,
                                        value)
        self._test_negative_set_and_get(oracledb.DB_TYPE_BLOB, temp_clob)
        self._test_negative_set_and_get(oracledb.DB_TYPE_NCLOB, temp_clob)
        value = 'Short CLOB value'
        self._test_positive_set_and_get(oracledb.DB_TYPE_CLOB, value, value)
        self._test_positive_set_and_get(oracledb.DB_TYPE_CLOB, None, None)
        self._test_negative_set_and_get(oracledb.DB_TYPE_CLOB, 5)

    def test_3715_DB_TYPE_NCLOB(self):
        "3715 - setting values on variables of type DB_TYPE_NCLOB"
        value = 'Short temp NCLOB value'
        temp_nclob = self.connection.createlob(oracledb.DB_TYPE_NCLOB)
        temp_nclob.write(value)
        self._test_positive_set_and_get(oracledb.DB_TYPE_NCLOB, temp_nclob,
                                        value)
        self._test_negative_set_and_get(oracledb.DB_TYPE_BLOB, temp_nclob)
        self._test_negative_set_and_get(oracledb.DB_TYPE_CLOB, temp_nclob)
        value = 'Short NCLOB Value'
        self._test_positive_set_and_get(oracledb.DB_TYPE_NCLOB, value, value)
        self._test_positive_set_and_get(oracledb.DB_TYPE_NCLOB, None, None)
        self._test_negative_set_and_get(oracledb.DB_TYPE_NCLOB, 5)

    def test_3716_DB_TYPE_BINARY_FLOAT(self):
        "3716 - setting values on variables of type DB_TYPE_BINARY_FLOAT"
        self._test_positive_set_and_get(oracledb.DB_TYPE_BINARY_FLOAT, 5, 5.0)
        self._test_positive_set_and_get(oracledb.DB_TYPE_BINARY_FLOAT,
                                        3.5, 3.5)
        self._test_positive_set_and_get(oracledb.DB_TYPE_BINARY_FLOAT,
                                        decimal.Decimal("24.5"), 24.5)
        self._test_positive_set_and_get(oracledb.DB_TYPE_BINARY_FLOAT,
                                        True, 1.0)
        self._test_positive_set_and_get(oracledb.DB_TYPE_BINARY_FLOAT,
                                        False, 0.0)
        self._test_positive_set_and_get(oracledb.DB_TYPE_BINARY_FLOAT,
                                        None, None)
        self._test_negative_set_and_get(oracledb.DB_TYPE_BINARY_FLOAT, "abc")

    def test_3717_DB_TYPE_BINARY_DOUBLE(self):
        "3717 - setting values on variables of type DB_TYPE_BINARY_DOUBLE"
        self._test_positive_set_and_get(oracledb.DB_TYPE_BINARY_DOUBLE, 5, 5.0)
        self._test_positive_set_and_get(oracledb.DB_TYPE_BINARY_DOUBLE,
                                        3.5, 3.5)
        self._test_positive_set_and_get(oracledb.DB_TYPE_BINARY_DOUBLE,
                                        decimal.Decimal("192.125"), 192.125)
        self._test_positive_set_and_get(oracledb.DB_TYPE_BINARY_DOUBLE,
                                        True, 1.0)
        self._test_positive_set_and_get(oracledb.DB_TYPE_BINARY_DOUBLE,
                                        False, 0.0)
        self._test_positive_set_and_get(oracledb.DB_TYPE_BINARY_DOUBLE,
                                        None, None)
        self._test_negative_set_and_get(oracledb.DB_TYPE_BINARY_DOUBLE, "abc")

    def test_3718_DB_TYPE_BOOLEAN(self):
        "3718 - setting values on variables of type DB_TYPE_BOOLEAN"
        self._test_positive_set_and_get(oracledb.DB_TYPE_BOOLEAN, 5, True)
        self._test_positive_set_and_get(oracledb.DB_TYPE_BOOLEAN, 2.0, True)
        self._test_positive_set_and_get(oracledb.DB_TYPE_BOOLEAN, "abc", True)
        self._test_positive_set_and_get(oracledb.DB_TYPE_BOOLEAN,
                                        decimal.Decimal("24.8"), True)
        self._test_positive_set_and_get(oracledb.DB_TYPE_BOOLEAN, 0.0, False)
        self._test_positive_set_and_get(oracledb.DB_TYPE_BOOLEAN, 0, False)
        self._test_positive_set_and_get(oracledb.DB_TYPE_BOOLEAN, None, None)

    def test_3719_DB_TYPE_INTERVAL_DS(self):
        "3719 - setting values on variables of type DB_TYPE_INTERVAL_DS"
        value = datetime.timedelta(days=5, seconds=56000, microseconds=123780)
        self._test_positive_set_and_get(oracledb.DB_TYPE_INTERVAL_DS,
                                        value, value)
        self._test_positive_set_and_get(oracledb.DB_TYPE_INTERVAL_DS,
                                        None, None)
        self._test_negative_set_and_get(oracledb.DB_TYPE_INTERVAL_DS, 5)

    def test_3720_DB_TYPE_ROWID(self):
        "3720 - setting values on variables of type DB_TYPE_ROWID"
        self._test_negative_set_and_get(oracledb.DB_TYPE_ROWID, 12345)
        self._test_negative_set_and_get(oracledb.DB_TYPE_ROWID, "523lkhlf")

    def test_3721_DB_TYPE_OBJECT(self):
        "3721 - setting values on variables of type DB_TYPE_OBJECT"
        obj_type = self.connection.gettype("UDT_OBJECT")
        obj = obj_type.newobject()
        plain_obj = self.get_db_object_as_plain_object(obj)
        self._test_positive_set_and_get(oracledb.DB_TYPE_OBJECT,
                                        obj, plain_obj, "UDT_OBJECT")
        self._test_positive_set_and_get(obj_type, obj, plain_obj)
        self._test_positive_set_and_get(oracledb.DB_TYPE_OBJECT,
                                        None, None, "UDT_OBJECT")
        self._test_positive_set_and_get(obj_type, None, None)
        self._test_negative_set_and_get(oracledb.DB_TYPE_OBJECT,
                                        "abc", "UDT_OBJECT")
        self._test_negative_set_and_get(oracledb.DB_TYPE_OBJECT, obj,
                                        "UDT_OBJECTARRAY")
        wrong_obj_type = self.connection.gettype("UDT_OBJECTARRAY")
        self._test_negative_set_and_get(wrong_obj_type, obj)

    @unittest.skipIf(test_env.get_client_version() < (21, 0),
                     "unsupported client")
    @unittest.skipIf(test_env.get_server_version() < (21, 0),
                     "unsupported server")
    def test_3722_DB_TYPE_JSON(self):
        "3722 - setting values on variables of type DB_TYPE_JSON"
        json_data = [
            5,
            25.25,
            decimal.Decimal("10.25"),
            True,
            False,
            datetime.datetime(2017, 5, 6),
            datetime.datetime(2017, 5, 6, 9, 36, 0, 300000),
            datetime.timedelta(days=5, seconds=56000, microseconds=123780),
            {},
            'String',
            b'Some bytes',
            {'keyA': 1, 'KeyB': 'Melbourne'},
            [],
            [1, "A"],
            {"name": None},
            {"name": "John"},
            {"age": 30},
            {"Permanent": True},
            {
                "employee": {
                    "name":"John",
                    "age": 30,
                    "city": "Delhi",
                    "Parmanent": True
                }
            },
            {
                "employees": ["John", "Matthew", "James"]
            },
            {
                "employees": [
                    {
                        "employee1": {"name": "John", "city": "Delhi"}
                    },
                    {
                        "employee2": {"name": "Matthew", "city": "Mumbai"}
                    },
                    {
                        "employee3": {"name": "James", "city": "Bangalore"}
                    }
                ]
            }
        ]
        self._test_positive_set_and_get(oracledb.DB_TYPE_JSON,
                                        json_data, json_data)
        self._test_positive_set_and_get(oracledb.DB_TYPE_JSON, None, None)

    def test_3723_DB_TYPE_CURSOR(self):
        "3723 - test setting values on variables of type DB_TYPE_CURSOR"
        cursor = self.connection.cursor()
        self._test_positive_set_and_get(oracledb.DB_TYPE_CURSOR, None, None)
        self._test_negative_set_and_get(oracledb.DB_TYPE_CURSOR, 5)

    def test_3724_fetch_null_column(self):
        "3724 - test fetching columns containing all null values"
        self.cursor.execute("""
                select
                    null,
                    to_char(null),
                    to_number(null),
                    to_date(null),
                    to_timestamp(null),
                    to_clob(null),
                    to_blob(null)
                from dual""")
        results = self.cursor.fetchall()
        self.assertEqual(results, [(None, None, None, None, None, None, None)])

    @unittest.skipIf(not test_env.get_is_thin(),
                     "thick mode doesn't support DB_TYPE_UROWID")
    def test_3725_DB_TYPE_UROWID(self):
        "3725 - setting values on variables of type DB_TYPE_UROWID"
        self._test_negative_set_and_get(oracledb.DB_TYPE_UROWID, 12345)
        self._test_negative_set_and_get(oracledb.DB_TYPE_UROWID, "523lkhlf")

    def test_3726_getvalue_with_index(self):
        "3726 - getting value with an specific index"
        var = self.cursor.var(oracledb.DB_TYPE_NUMBER, 1000, 2)
        var.setvalue(0, 10)
        self.assertEqual(var.getvalue(0), 10)
        self.assertEqual(var.getvalue(1), None)
        self.assertRaises(IndexError, var.getvalue, 4)

    def test_3727_get_buffer_size(self):
        "3727 - getting buffer_size attribute"
        test_values = [
            (oracledb.DB_TYPE_NUMBER, 200, 22),
            (oracledb.DB_TYPE_VARCHAR, 3000, 12000),
            (oracledb.DB_TYPE_RAW, 4000, 4000),
            (oracledb.DB_TYPE_NCHAR, 1000, 4000),
            (oracledb.DB_TYPE_CHAR, 2000, 8000)
        ]
        for typ, size, buffer_size in test_values:
            var = self.cursor.var(typ, size)
            self.assertEqual(var.buffer_size, buffer_size)

if __name__ == "__main__":
    test_env.run_test_cases()



#------------------------------------------------------------------------------
# Copyright (c) 2020, 2022, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# connection.pyx
#
# Cython file defining the base Connection implementation class (embedded in
# base_impl.pyx).
#------------------------------------------------------------------------------

cdef class BaseConnImpl:

    def __init__(self, str dsn, ConnectParamsImpl params):
        self.dsn = dsn
        self.username = params.user

    cdef object _check_value(self, DbType dbtype, BaseDbObjectTypeImpl objtype,
                             object value, bint* is_ok):
        """
        Checks that the specified Python value is acceptable for the given
        database type. If the "is_ok" parameter is passed as NULL, an exception
        is raised.  The value to use is returned (possibly modified from the
        value passed in).
        """
        cdef:
            uint32_t db_type_num
            BaseLobImpl lob_impl

        # null values are always accepted
        if value is None:
            return value

        # check to see if the Python value is accepted and perform any
        # necessary adjustments
        db_type_num = dbtype.num
        if db_type_num in (DB_TYPE_NUM_NUMBER,
                           DB_TYPE_NUM_BINARY_INTEGER,
                           DB_TYPE_NUM_BINARY_DOUBLE,
                           DB_TYPE_NUM_BINARY_FLOAT):
            if isinstance(value, (PY_TYPE_BOOL, int, float, PY_TYPE_DECIMAL)):
                if db_type_num in (DB_TYPE_NUM_BINARY_FLOAT,
                                   DB_TYPE_NUM_BINARY_DOUBLE):
                    return float(value)
                elif db_type_num == DB_TYPE_NUM_BINARY_INTEGER \
                        or cpython.PyBool_Check(value):
                    return int(value)
                return value
        elif db_type_num in (DB_TYPE_NUM_CHAR,
                             DB_TYPE_NUM_VARCHAR,
                             DB_TYPE_NUM_NCHAR,
                             DB_TYPE_NUM_NVARCHAR,
                             DB_TYPE_NUM_LONG_VARCHAR,
                             DB_TYPE_NUM_LONG_NVARCHAR):
            if isinstance(value, bytes):
                return (<bytes> value).decode()
            elif isinstance(value, str):
                return value
        elif db_type_num in (DB_TYPE_NUM_RAW, DB_TYPE_NUM_LONG_RAW):
            if isinstance(value, str):
                return (<str> value).encode()
            elif isinstance(value, bytes):
                return value
        elif db_type_num in (DB_TYPE_NUM_DATE,
                             DB_TYPE_NUM_TIMESTAMP,
                             DB_TYPE_NUM_TIMESTAMP_LTZ,
                             DB_TYPE_NUM_TIMESTAMP_TZ):
            if cydatetime.PyDateTime_Check(value) \
                    or cydatetime.PyDate_Check(value):
                return value
        elif db_type_num == DB_TYPE_NUM_INTERVAL_DS:
            if isinstance(value, PY_TYPE_TIMEDELTA):
                return value
        elif db_type_num in (DB_TYPE_NUM_CLOB,
                             DB_TYPE_NUM_NCLOB,
                             DB_TYPE_NUM_BLOB):
            if isinstance(value, PY_TYPE_LOB):
                lob_impl = value._impl
                if lob_impl.dbtype is not dbtype:
                    if is_ok != NULL:
                        is_ok[0] = False
                        return value
                    errors._raise_err(errors.ERR_LOB_OF_WRONG_TYPE,
                                      actual_type_name=lob_impl.dbtype.name,
                                      expected_type_name=dbtype.name)
                return value
            elif isinstance(value, (bytes, str)):
                if db_type_num == DB_TYPE_NUM_BLOB:
                    if isinstance(value, str):
                        value = value.encode()
                elif isinstance(value, bytes):
                    value = value.decode()
                lob_impl = self.create_temp_lob_impl(dbtype)
                if value:
                    lob_impl.write(value, 1)
                return PY_TYPE_LOB._from_impl(lob_impl)
        elif db_type_num == DB_TYPE_NUM_OBJECT:
            if isinstance(value, PY_TYPE_DB_OBJECT):
                if value._impl.type != objtype:
                    if is_ok != NULL:
                        is_ok[0] = False
                        return value
                    errors._raise_err(errors.ERR_WRONG_OBJECT_TYPE,
                                      actual_schema=value.type.schema,
                                      actual_name=value.type.name,
                                      expected_schema=objtype.schema,
                                      expected_name=objtype.name)
                return value
        elif db_type_num == DB_TYPE_NUM_CURSOR:
            if isinstance(value, PY_TYPE_CURSOR):
                return value
        elif db_type_num == DB_TYPE_NUM_BOOLEAN:
            return bool(value)
        elif db_type_num == DB_TYPE_NUM_JSON:
            return value
        else:
            if is_ok != NULL:
                is_ok[0] = False
                return value
            errors._raise_err(errors.ERR_UNSUPPORTED_TYPE_SET,
                              db_type_name=dbtype.name)

        # the Python value was not considered acceptable
        if is_ok != NULL:
            is_ok[0] = False
            return value
        errors._raise_err(errors.ERR_UNSUPPORTED_PYTHON_TYPE_FOR_DB_TYPE,
                          py_type_name=type(value).__name__,
                          db_type_name=dbtype.name)

    @utils.CheckImpls("getting a connection OCI attribute")
    def _get_oci_attr(self, uint32_t handle_type, uint32_t attr_num,
                      uint32_t attr_type):
        pass

    @utils.CheckImpls("setting a connection OCI attribute")
    def _set_oci_attr(self, uint32_t handle_type, uint32_t attr_num,
                      uint32_t attr_type, object value):
        pass

    @utils.CheckImpls("aborting a currently executing statement")
    def cancel(self):
        pass

    @utils.CheckImpls("changing a password")
    def change_password(self, old_password, new_password):
        pass

    @utils.CheckImpls("checking if the connection is healthy")
    def get_is_healthy(self):
        pass

    @utils.CheckImpls("closing a connection")
    def close(self, in_del=False):
        pass

    @utils.CheckImpls("committing a transaction")
    def commit(self):
        pass

    @utils.CheckImpls("creating a cursor")
    def create_cursor_impl(self):
        pass

    @utils.CheckImpls("creating a queue")
    def create_queue_impl(self):
        pass

    @utils.CheckImpls("creating a SODA database object")
    def create_soda_database_impl(self, conn):
        pass

    @utils.CheckImpls("creating a subscription")
    def create_subscr_impl(self, object conn, object callback,
                           uint32_t namespace, str name, uint32_t protocol,
                           str ip_address, uint32_t port, uint32_t timeout,
                           uint32_t operations, uint32_t qos,
                           uint8_t grouping_class, uint32_t grouping_value,
                           uint8_t grouping_type, bint client_initiated):
        pass

    @utils.CheckImpls("creating a temporary LOB")
    def create_temp_lob_impl(self, DbType dbtype):
        pass

    @utils.CheckImpls("getting the call timeout")
    def get_call_timeout(self):
        pass

    @utils.CheckImpls("getting the current schema")
    def get_current_schema(self):
        pass

    @utils.CheckImpls("getting the edition")
    def get_edition(self):
        pass

    @utils.CheckImpls("getting the external name")
    def get_external_name(self):
        pass

    @utils.CheckImpls("getting the OCI service context handle")
    def get_handle(self):
        pass

    @utils.CheckImpls("getting the internal name")
    def get_internal_name(self):
        pass

    @utils.CheckImpls("getting the logical transaction id")
    def get_ltxid(self):
        pass

    @utils.CheckImpls("getting the statement cache size")
    def get_stmt_cache_size(self):
        pass

    @utils.CheckImpls("getting an object type")
    def get_type(self, object conn, str name):
        pass

    @utils.CheckImpls("getting the database version")
    def get_version(self):
        pass

    @utils.CheckImpls("pinging the database")
    def ping(self):
        pass

    @utils.CheckImpls("rolling back a transaction")
    def rollback(self):
        pass

    @utils.CheckImpls("setting the action")
    def set_action(self, value):
        pass

    @utils.CheckImpls("setting the call timeout")
    def set_call_timeout(self, value):
        pass

    @utils.CheckImpls("setting the client identifier")
    def set_client_identifier(self, value):
        pass

    @utils.CheckImpls("setting the client info")
    def set_client_info(self, value):
        pass

    @utils.CheckImpls("setting the current schema")
    def set_current_schema(self, value):
        pass

    @utils.CheckImpls("setting the database operation")
    def set_dbop(self, value):
        pass

    @utils.CheckImpls("setting the execution context id")
    def set_econtext_id(self, value):
        pass

    @utils.CheckImpls("setting the external name")
    def set_external_name(self, value):
        pass

    @utils.CheckImpls("setting the internal name")
    def set_internal_name(self, value):
        pass

    @utils.CheckImpls("setting the module")
    def set_module(self, value):
        pass

    @utils.CheckImpls("setting the statement cache size")
    def set_stmt_cache_size(self, value):
        pass

    @utils.CheckImpls("shutting down the database")
    def shutdown(self, uint32_t mode):
        pass

    @utils.CheckImpls("starting up the database")
    def startup(self, bint force, bint restrict, str pfile):
        pass

    @utils.CheckImpls("starting a TPC (two-phase commit) transaction")
    def tpc_begin(self, xid, uint32_t flags, uint32_t timeout):
        pass

    @utils.CheckImpls("committing a TPC (two-phase commit) transaction")
    def tpc_commit(self, xid, bint one_phase):
        pass

    @utils.CheckImpls("ending a TPC (two-phase commit) transaction")
    def tpc_end(self, xid, uint32_t flags):
        pass

    @utils.CheckImpls("forgetting a TPC (two-phase commit) transaction")
    def tpc_forget(self, xid):
        pass

    @utils.CheckImpls("preparing a TPC (two-phase commit) transaction")
    def tpc_prepare(self, xid):
        pass

    @utils.CheckImpls("rolling back a TPC (two-phase commit) transaction")
    def tpc_rollback(self, xid):
        pass



#------------------------------------------------------------------------------
# Copyright (c) 2020, 2023, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# Sets the environment used by the python-oracledb test suite. Production
# applications should consider using External Authentication to avoid hard
# coded credentials.
#
# You can set values in environment variables to bypass having the test suite
# request the information it requires.
#
#   PYO_TEST_MAIN_USER: user used for most test cases
#   PYO_TEST_MAIN_PASSWORD: password of user used for most test cases
#   PYO_TEST_PROXY_USER: user for testing proxying
#   PYO_TEST_PROXY_PASSWORD: password of user for testing proxying
#   PYO_TEST_CONNECT_STRING: connect string for test suite
#   PYO_TEST_ADMIN_USER: administrative user for test suite
#   PYO_TEST_ADMIN_PASSWORD: administrative password for test suite
#   PYO_TEST_WALLET_LOCATION: location of wallet file (thin mode, mTLS)
#   PYO_TEST_WALLET_PASSWORD: password for wallet file (thin mode, mTLS)
#   PYO_TEST_DRIVER_MODE: python-oracledb mode (thick or thin) to use
#   PYO_TEST_EXTERNAL_USER: user for testing external authentication
#
# PYO_TEST_CONNECT_STRING can be set to an Easy Connect string, or a
# Net Service Name from a tnsnames.ora file or external naming service,
# or it can be the name of a local Oracle database instance.
#
# If oracledb is using Instant Client, then an Easy Connect string is generally
# appropriate. The syntax is:
#
#   [//]host_name[:port][/service_name][:server_type][/instance_name]
#
# Commonly just the host_name and service_name are needed
# e.g. "localhost/orclpdb1" or "localhost/XEPDB1"
#
# If using a tnsnames.ora file, the file can be in a default
# location such as $ORACLE_HOME/network/admin/tnsnames.ora or
# /etc/tnsnames.ora.  Alternatively set the TNS_ADMIN environment
# variable and put the file in $TNS_ADMIN/tnsnames.ora.
#
# The administrative user for cloud databases is ADMIN and the administrative
# user for on premises databases is SYSTEM.
#------------------------------------------------------------------------------

import getpass
import os
import sys
import unittest

import oracledb

# default values
DEFAULT_MAIN_USER = "pythontest"
DEFAULT_PROXY_USER = "pythontestproxy"
DEFAULT_CONNECT_STRING = "localhost/orclpdb1"

# dictionary containing all parameters; these are acquired as needed by the
# methods below (which should be used instead of consulting this dictionary
# directly) and then stored so that a value is not requested more than once
PARAMETERS = {}

def get_value(name, label, default_value=None, password=False):
    try:
        return PARAMETERS[name]
    except KeyError:
        pass
    env_name = "PYO_TEST_" + name
    value = os.environ.get(env_name)
    if value is None:
        if default_value is not None:
            label += " [%s]" % default_value
        label += ": "
        if password:
            value = getpass.getpass(label)
        else:
            value = input(label).strip()
    if not value:
        value = default_value
    PARAMETERS[name] = value
    return value

def get_admin_connection():
    admin_user = get_value("ADMIN_USER", "Administrative user", "admin")
    admin_password = get_value("ADMIN_PASSWORD", f"Password for {admin_user}",
                               password=True)
    params = get_connect_params()
    if admin_user and admin_user.upper() == "SYS":
        params = params.copy()
        params.set(mode=oracledb.AUTH_MODE_SYSDBA)
    return oracledb.connect(dsn=get_connect_string(), params=params,
                            user=admin_user, password=admin_password)

def get_charset_ratios():
    value = PARAMETERS.get("CS_RATIO")
    if value is None:
        connection = get_connection()
        cursor = connection.cursor()
        cursor.execute("""
                select
                    cast('X' as varchar2(1)),
                    cast('Y' as nvarchar2(1))
                from dual""")
        varchar_column_info, nvarchar_column_info = cursor.description
        value = (varchar_column_info[3], nvarchar_column_info[3])
        PARAMETERS["CS_RATIO"] = value
    return value

def get_client_version():
    name = "CLIENT_VERSION"
    value = PARAMETERS.get(name)
    if value is None:
        if get_is_thin():
            value = (21, 3)
        else:
            value = oracledb.clientversion()[:2]
        PARAMETERS[name] = value
    return value

def get_connect_params():
    name = "CONNECT_PARAMS"
    params = PARAMETERS.get(name)
    if params is None:
        wallet_location = get_wallet_location()
        params = oracledb.ConnectParams(user=get_main_user(),
                                        password=get_main_password(),
                                        config_dir=wallet_location,
                                        wallet_location=wallet_location,
                                        wallet_password=get_wallet_password())
        PARAMETERS[name] = params
    return params

def get_connection(dsn=None, **kwargs):
    if dsn is None:
        dsn = get_connect_string()
    return oracledb.connect(dsn=dsn, params=get_connect_params(), **kwargs)

def get_connect_string():
    return get_value("CONNECT_STRING", "Connect String",
                     DEFAULT_CONNECT_STRING)

def get_is_thin():
    driver_mode = get_value("DRIVER_MODE", "Driver mode (thin|thick)", "thin")
    return driver_mode == "thin"

def get_main_password():
    return get_value("MAIN_PASSWORD", f"Password for {get_main_user()}",
                     password=True)

def get_main_user():
    return get_value("MAIN_USER", "Main User Name", DEFAULT_MAIN_USER)

def get_pool(**kwargs):
    return oracledb.create_pool(dsn=get_connect_string(),
                                params=get_pool_params(), **kwargs)

def get_pool_params():
    name = "POOL_PARAMS"
    params = PARAMETERS.get(name)
    if params is None:
        wallet_location = get_wallet_location()
        params = oracledb.PoolParams(user=get_main_user(),
                                     password=get_main_password(),
                                     config_dir=wallet_location,
                                     wallet_location=wallet_location,
                                     wallet_password=get_wallet_password())
        PARAMETERS[name] = params
    return params

def get_proxy_password():
    return get_value("PROXY_PASSWORD", f"Password for {get_proxy_user()}",
                     password=True)

def get_proxy_user():
    return get_value("PROXY_USER", "Proxy User Name", DEFAULT_PROXY_USER)

def get_sleep_proc_name():
    server_version = get_server_version()
    return "dbms_session.sleep" if server_version[0] >= 18 \
            else "dbms_lock.sleep"

def get_server_version():
    name = "SERVER_VERSION"
    value = PARAMETERS.get(name)
    if value is None:
        conn = get_connection()
        value = tuple(int(s) for s in conn.version.split("."))[:2]
        PARAMETERS[name] = value
    return value

def get_wallet_location():
    if get_is_thin():
        return get_value("WALLET_LOCATION", "Wallet Location")

def get_wallet_password():
    if get_is_thin():
        return get_value("WALLET_PASSWORD", "Wallet Password", password=True)

def get_external_user():
    if not get_is_thin():
        return get_value("EXTERNAL_USER", "External User")

def is_on_oracle_cloud(connection):
    server = get_server_version()
    if server < (18, 0):
        return False
    cursor = connection.cursor()
    cursor.execute("""
            select sys_context('userenv', 'cloud_service')
            from dual""")
    service_name, = cursor.fetchone()
    return service_name is not None

def run_sql_script(conn, script_name, **kwargs):
    statement_parts = []
    cursor = conn.cursor()
    replace_values = [("&" + k + ".", v) for k, v in kwargs.items()] + \
                     [("&" + k, v) for k, v in kwargs.items()]
    script_dir = os.path.dirname(os.path.abspath(sys.argv[0]))
    file_name = os.path.join(script_dir, "sql", script_name + ".sql")
    for line in open(file_name):
        if line.strip() == "/":
            statement = "".join(statement_parts).strip()
            if statement:
                for search_value, replace_value in replace_values:
                    statement = statement.replace(search_value, replace_value)
                try:
                    cursor.execute(statement)
                except:
                    print("Failed to execute SQL:", statement)
                    raise
            statement_parts = []
        else:
            statement_parts.append(line)
    cursor.execute("""
            select name, type, line, position, text
            from dba_errors
            where owner = upper(:owner)
            order by name, type, line, position""",
            owner = get_main_user())
    prev_name = prev_obj_type = None
    for name, obj_type, line_num, position, text in cursor:
        if name != prev_name or obj_type != prev_obj_type:
            print("%s (%s)" % (name, obj_type))
            prev_name = name
            prev_obj_type = obj_type
        print("    %s/%s %s" % (line_num, position, text))

def run_test_cases():
    unittest.main(testRunner=unittest.TextTestRunner(verbosity=2))

def skip_soda_tests():
    if get_is_thin():
        return True
    client = get_client_version()
    if client < (18, 3):
        return True
    server = get_server_version()
    if server < (18, 0):
        return True
    if server > (20, 1) and client < (20, 1):
        return True
    return False


class FetchLobsContextManager:

    def __init__(self, desired_value):
        self.desired_value = desired_value

    def __enter__(self):
        self.original_value = oracledb.defaults.fetch_lobs
        oracledb.defaults.fetch_lobs = self.desired_value
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        oracledb.defaults.fetch_lobs = self.original_value


class SystemStatInfo:
    stat_name = None

    def __init__(self, connection):
        self.prev_value = 0
        self.admin_conn = get_admin_connection()
        with connection.cursor() as cursor:
            cursor.execute("select sys_context('userenv', 'sid') from dual")
            self.sid, = cursor.fetchone()
        self.get_value()

    def get_value(self):
        with self.admin_conn.cursor() as cursor:
            cursor.execute("""
                    select ss.value
                    from v$sesstat ss, v$statname sn
                    where ss.sid = :sid
                        and ss.statistic# = sn.statistic#
                        and sn.name = :stat_name""",
                    sid=self.sid, stat_name=self.stat_name)
            current_value, = cursor.fetchone()
            diff_value = current_value - self.prev_value
            self.prev_value = current_value
            return diff_value


class RoundTripInfo(SystemStatInfo):
    stat_name = "SQL*Net roundtrips to/from client"


class ParseCountInfo(SystemStatInfo):
    stat_name = "parse count (total)"


class BaseTestCase(unittest.TestCase):
    requires_connection = True

    def assertParseCount(self, n):
        self.assertEqual(self.parse_count_info.get_value(), n)

    def assertRoundTrips(self, n):
        self.assertEqual(self.round_trip_info.get_value(), n)

    def get_and_clear_queue(self, queue_name, payload_type=None,
                            message="not supported with this client/server " \
                                    "combination"):
        if payload_type == "JSON":
            if get_client_version() < (21, 0) \
                    or get_server_version() < (21, 0):
                self.skipTest(message)
        elif isinstance(payload_type, str):
            payload_type = self.connection.gettype(payload_type)
        queue = self.connection.queue(queue_name, payload_type)
        queue.deqoptions.wait = oracledb.DEQ_NO_WAIT
        queue.deqoptions.deliverymode = oracledb.MSG_PERSISTENT_OR_BUFFERED
        queue.deqoptions.visibility = oracledb.DEQ_IMMEDIATE
        while queue.deqone():
            pass
        return self.connection.queue(queue_name, payload_type)

    def get_db_object_as_plain_object(self, obj):
        if obj.type.iscollection:
            element_values = []
            for value in obj.aslist():
                if isinstance(value, oracledb.DbObject):
                    value = self.get_db_object_as_plain_object(value)
                elif isinstance(value, oracledb.LOB):
                    value = value.read()
                element_values.append(value)
            return element_values
        attr_values = []
        for attribute in obj.type.attributes:
            value = getattr(obj, attribute.name)
            if isinstance(value, oracledb.DbObject):
                value = self.get_db_object_as_plain_object(value)
            elif isinstance(value, oracledb.LOB):
                value = value.read()
            attr_values.append(value)
        return tuple(attr_values)

    def get_soda_database(self, minclient=(18, 3), minserver=(18, 0),
                          message="not supported with this client/server " \
                                  "combination"):
        client = get_client_version()
        if client < minclient:
            self.skipTest(message)
        server = get_server_version()
        if server < minserver:
            self.skipTest(message)
        if server > (20, 1) and client < (20, 1):
            self.skipTest(message)
        return self.connection.getSodaDatabase()

    def is_on_oracle_cloud(self, connection=None):
        if connection is None:
            connection = self.connection
        return is_on_oracle_cloud(connection)

    def setUp(self):
        if self.requires_connection:
            self.connection = get_connection()
            self.cursor = self.connection.cursor()
            self.cursor.execute("alter session set time_zone = '+00:00'")

    def setup_parse_count_checker(self):
        self.parse_count_info = ParseCountInfo(self.connection)

    def setup_round_trip_checker(self):
        self.round_trip_info = RoundTripInfo(self.connection)

    def tearDown(self):
        if self.requires_connection:
            self.connection.close()
            del self.cursor
            del self.connection

# ensure that thick mode is enabled, if desired
if not get_is_thin():
    oracledb.init_oracle_client()



#------------------------------------------------------------------------------
# Copyright (c) 2020, 2022, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

"""
2200 - Module for testing number variables
"""

import decimal

import oracledb
import test_env

class TestCase(test_env.BaseTestCase):

    def output_type_handler_binary_int(self, cursor, name, default_type, size,
                                       precision, scale):
        return cursor.var(oracledb.DB_TYPE_BINARY_INTEGER,
                          arraysize=cursor.arraysize)

    def output_type_handler_decimal(self, cursor, name, default_type, size,
                                    precision, scale):
        if default_type == oracledb.NUMBER:
            return cursor.var(str, 255, outconverter=decimal.Decimal,
                              arraysize=cursor.arraysize)

    def output_type_handler_str(self, cursor, name, default_type, size,
                                precision, scale):
        return cursor.var(str, 255, arraysize=cursor.arraysize)


    def setUp(self):
        super().setUp()
        self.raw_data = []
        self.data_by_key = {}
        for i in range(1, 11):
            number_col = i + i * 0.25
            float_col = i + i * 0.75
            unconstrained_col = i ** 3 + i * 0.5
            if i % 2:
                nullable_col = 143 ** i
            else:
                nullable_col = None
            data_tuple = (i, 38 ** i, number_col, float_col,
                          unconstrained_col, nullable_col)
            self.raw_data.append(data_tuple)
            self.data_by_key[i] = data_tuple

    def test_2200_bind_boolean(self):
        "2200 - test binding in a boolean"
        result = self.cursor.callfunc("pkg_TestBooleans.GetStringRep", str,
                                      (True,))
        self.assertEqual(result, "TRUE")

    def test_2201_bind_boolean_as_number(self):
        "2201 - test binding in a boolean as a number"
        var = self.cursor.var(oracledb.NUMBER)
        var.setvalue(0, True)
        self.cursor.execute("select :1 from dual", [var])
        result, = self.cursor.fetchone()
        self.assertEqual(result, 1)
        var.setvalue(0, False)
        self.cursor.execute("select :1 from dual", [var])
        result, = self.cursor.fetchone()
        self.assertEqual(result, 0)

    def test_2202_bind_decimal(self):
        "2202 - test binding in a decimal.Decimal"
        self.cursor.execute("""
                select * from TestNumbers
                where NumberCol - :value1 - :value2 = trunc(NumberCol)""",
                value1=decimal.Decimal("0.20"),
                value2=decimal.Decimal("0.05"))
        expected_data = [self.data_by_key[1], self.data_by_key[5],
                         self.data_by_key[9]]
        self.assertEqual(self.cursor.fetchall(), expected_data)

    def test_2203_bind_float(self):
        "2203 - test binding in a float"
        self.cursor.execute("""
                select * from TestNumbers
                where NumberCol - :value = trunc(NumberCol)""",
                value=0.25)
        expected_data = [self.data_by_key[1], self.data_by_key[5],
                         self.data_by_key[9]]
        self.assertEqual(self.cursor.fetchall(), expected_data)

    def test_2204_bind_integer(self):
        "2204 - test binding in an integer"
        self.cursor.execute("""
                select * from TestNumbers
                where IntCol = :value""",
                value = 2)
        self.assertEqual(self.cursor.fetchall(), [self.data_by_key[2]])

    def test_2205_bind_large_long_as_oracle_number(self):
        "2205 - test binding in a large long integer as Oracle number"
        in_val = 6088343244
        value_var = self.cursor.var(oracledb.NUMBER)
        value_var.setvalue(0, in_val)
        self.cursor.execute("""
                begin
                    :value := :value + 5;
                end;""",
                value=value_var)
        value = value_var.getvalue()
        self.assertEqual(value, in_val + 5)

    def test_2206_bind_large_long_as_integer(self):
        "2206 - test binding in a large long integer as Python integer"
        long_value = -9999999999999999999
        self.cursor.execute("select :value from dual", value=long_value)
        result, = self.cursor.fetchone()
        self.assertEqual(result, long_value)

    def test_2207_bind_integer_after_string(self):
        "2207 - test binding in an integer after setting input sizes to string"
        self.cursor.setinputsizes(value=15)
        self.cursor.execute("""
                select * from TestNumbers
                where IntCol = :value""",
                value=3)
        self.assertEqual(self.cursor.fetchall(), [self.data_by_key[3]])

    def test_2208_bind_decimal_after_number(self):
        "2208 - test binding in a decimal after setting input sizes to number"
        cursor = self.connection.cursor()
        value = decimal.Decimal("319438950232418390.273596")
        cursor.setinputsizes(value=oracledb.NUMBER)
        cursor.outputtypehandler = self.output_type_handler_decimal
        cursor.execute("select :value from dual", value=value)
        out_value, = cursor.fetchone()
        self.assertEqual(out_value, value)

    def test_2209_bind_null(self):
        "2209 - test binding in a null"
        self.cursor.execute("""
                select * from TestNumbers
                where IntCol = :value""",
                value=None)
        self.assertEqual(self.cursor.fetchall(), [])

    def test_2210_bind_number_array_direct(self):
        "2210 - test binding in a number array"
        return_value = self.cursor.var(oracledb.NUMBER)
        array = [r[2] for r in self.raw_data]
        statement = """
                begin
                    :return_value := pkg_TestNumberArrays.TestInArrays(
                        :start_value, :array);
                end;"""
        self.cursor.execute(statement, return_value=return_value,
                            start_value=5, array=array)
        self.assertEqual(return_value.getvalue(), 73.75)
        array = list(range(15))
        self.cursor.execute(statement, start_value=10, array=array)
        self.assertEqual(return_value.getvalue(), 115.0)

    def test_2211_bind_number_array_by_sizes(self):
        "2211 - test binding in a number array (with setinputsizes)"
        return_value = self.cursor.var(oracledb.NUMBER)
        self.cursor.setinputsizes(array = [oracledb.NUMBER, 10])
        array = [r[2] for r in self.raw_data]
        self.cursor.execute("""
                begin
                    :return_value := pkg_TestNumberArrays.TestInArrays(
                        :start_value, :array);
                end;""",
                return_value=return_value,
                start_value=6,
                array=array)
        self.assertEqual(return_value.getvalue(), 74.75)

    def test_2212_bind_number_array_by_var(self):
        "2212 - test binding in a number array (with arrayvar)"
        return_value = self.cursor.var(oracledb.NUMBER)
        array = self.cursor.arrayvar(oracledb.NUMBER,
                                     [r[2] for r in self.raw_data])
        self.cursor.execute("""
                begin
                    :return_value := pkg_TestNumberArrays.TestInArrays(
                        :integer_value, :array);
                end;""",
                return_value = return_value,
                integer_value = 7,
                array = array)
        self.assertEqual(return_value.getvalue(), 75.75)

    def test_2213_bind_zero_length_number_array_by_var(self):
        "2213 - test binding in a zero length number array (with arrayvar)"
        return_value = self.cursor.var(oracledb.NUMBER)
        array = self.cursor.arrayvar(oracledb.NUMBER, 0)
        self.cursor.execute("""
                begin
                    :return_value := pkg_TestNumberArrays.TestInArrays(
                        :integer_value, :array);
                end;""",
                return_value=return_value,
                integer_value=8,
                array=array)
        self.assertEqual(return_value.getvalue(), 8.0)
        self.assertEqual(array.getvalue(), [])

    def test_2214_bind_in_out_number_array_by_var(self):
        "2214 - test binding in/out a number array (with arrayvar)"
        array = self.cursor.arrayvar(oracledb.NUMBER, 10)
        original_data = [r[2] for r in self.raw_data]
        expected_data = [original_data[i - 1] * 10 for i in range(1, 6)] + \
                         original_data[5:]
        array.setvalue(0, original_data)
        self.cursor.execute("""
                begin
                    pkg_TestNumberArrays.TestInOutArrays(:num_elems, :array);
                end;""",
                num_elems=5,
                array=array)
        self.assertEqual(array.getvalue(), expected_data)

    def test_2215_bind_out_number_array_by_var(self):
        "2215 - test binding out a Number array (with arrayvar)"
        array = self.cursor.arrayvar(oracledb.NUMBER, 6)
        expected_data = [i * 100 for i in range(1, 7)]
        self.cursor.execute("""
                begin
                    pkg_TestNumberArrays.TestOutArrays(:num_elems, :array);
                end;""",
                num_elems=6,
                array=array)
        self.assertEqual(array.getvalue(), expected_data)

    def test_2216_bind_out_set_input_sizes(self):
        "2216 - test binding out with set input sizes defined"
        bind_vars = self.cursor.setinputsizes(value = oracledb.NUMBER)
        self.cursor.execute("""
                begin
                    :value := 5;
                end;""")
        self.assertEqual(bind_vars["value"].getvalue(), 5)

    def test_2217_bind_in_out_set_input_sizes(self):
        "2217 - test binding in/out with set input sizes defined"
        bind_vars = self.cursor.setinputsizes(value = oracledb.NUMBER)
        self.cursor.execute("""
                begin
                    :value := :value + 5;
                end;""",
                value = 1.25)
        self.assertEqual(bind_vars["value"].getvalue(), 6.25)

    def test_2218_bind_out_var(self):
        "2218 - test binding out with cursor.var() method"
        var = self.cursor.var(oracledb.NUMBER)
        self.cursor.execute("""
                begin
                    :value := 5;
                end;""",
                value = var)
        self.assertEqual(var.getvalue(), 5)

    def test_2219_bind_in_out_var_direct_set(self):
        "2219 - test binding in/out with cursor.var() method"
        var = self.cursor.var(oracledb.NUMBER)
        var.setvalue(0, 2.25)
        self.cursor.execute("""
                begin
                    :value := :value + 5;
                end;""",
                value = var)
        self.assertEqual(var.getvalue(), 7.25)

    def test_2220_cursor_description(self):
        "2220 - test cursor description is accurate"
        self.cursor.execute("select * from TestNumbers")
        expected_value = [
            ('INTCOL', oracledb.DB_TYPE_NUMBER, 10, None, 9, 0, False),
            ('LONGINTCOL', oracledb.DB_TYPE_NUMBER, 17, None, 16, 0, False),
            ('NUMBERCOL', oracledb.DB_TYPE_NUMBER, 13, None, 9, 2, False),
            ('FLOATCOL', oracledb.DB_TYPE_NUMBER, 127, None, 126, -127, False),
            ('UNCONSTRAINEDCOL', oracledb.DB_TYPE_NUMBER, 127, None, 0, -127,
                    False),
            ('NULLABLECOL', oracledb.DB_TYPE_NUMBER, 39, None, 38, 0, True)
        ]
        self.assertEqual(self.cursor.description, expected_value)

    def test_2221_fetchall(self):
        "2221 - test that fetching all of the data returns the correct results"
        self.cursor.execute("select * From TestNumbers order by IntCol")
        self.assertEqual(self.cursor.fetchall(), self.raw_data)
        self.assertEqual(self.cursor.fetchall(), [])

    def test_2222_fetchmany(self):
        "2222 - test that fetching data in chunks returns the correct results"
        self.cursor.execute("select * From TestNumbers order by IntCol")
        self.assertEqual(self.cursor.fetchmany(3), self.raw_data[0:3])
        self.assertEqual(self.cursor.fetchmany(2), self.raw_data[3:5])
        self.assertEqual(self.cursor.fetchmany(4), self.raw_data[5:9])
        self.assertEqual(self.cursor.fetchmany(3), self.raw_data[9:])
        self.assertEqual(self.cursor.fetchmany(3), [])

    def test_2223_fetchone(self):
        "2223 - test that fetching a single row returns the correct results"
        self.cursor.execute("""
                select *
                from TestNumbers
                where IntCol in (3, 4)
                order by IntCol""")
        self.assertEqual(self.cursor.fetchone(), self.data_by_key[3])
        self.assertEqual(self.cursor.fetchone(), self.data_by_key[4])
        self.assertEqual(self.cursor.fetchone(), None)

    def test_2224_return_as_long(self):
        "2224 - test that fetching a long integer returns such in Python"
        self.cursor.execute("""
                select NullableCol
                from TestNumbers
                where IntCol = 9""")
        col, = self.cursor.fetchone()
        self.assertEqual(col, 25004854810776297743)

    def test_2225_return_constant_float(self):
        "2225 - test fetching a floating point number returns such in Python"
        self.cursor.execute("select 1.25 from dual")
        result, = self.cursor.fetchone()
        self.assertEqual(result, 1.25)

    def test_2226_return_constant_integer(self):
        "2226 - test that fetching an integer returns such in Python"
        self.cursor.execute("select 148 from dual")
        result, = self.cursor.fetchone()
        self.assertEqual(result, 148)
        self.assertTrue(isinstance(result, int), "integer not returned")

    def test_2227_acceptable_boundary_numbers(self):
        "2227 - test that acceptable boundary numbers are handled properly"
        in_values = [decimal.Decimal("9.99999999999999e+125"),
                     decimal.Decimal("-9.99999999999999e+125"), 0.0, 1e-130,
                     -1e-130]
        out_values = [int("9" * 15 + "0" * 111), -int("9" * 15 + "0" * 111),
                      0, 1e-130, -1e-130]
        for in_value, out_value in zip(in_values, out_values):
            self.cursor.execute("select :1 from dual", (in_value,))
            result, = self.cursor.fetchone()
            self.assertEqual(result, out_value)

    def test_2228_unacceptable_boundary_numbers(self):
        "2228 - test that unacceptable boundary numbers are rejected"
        in_values = [1e126, -1e126, float("inf"), float("-inf"),
                     float("NaN"), decimal.Decimal("1e126"),
                     decimal.Decimal("-1e126"), decimal.Decimal("inf"),
                     decimal.Decimal("-inf"), decimal.Decimal("NaN")]
        no_rep_err = "^DPY-4003:"
        invalid_err = "^DPY-4004:"
        expected_errors = [no_rep_err, no_rep_err, invalid_err, invalid_err,
                           invalid_err, no_rep_err, no_rep_err, invalid_err,
                           invalid_err, invalid_err]
        for in_value, error in zip(in_values, expected_errors):
            self.assertRaisesRegex(oracledb.DatabaseError, error,
                                   self.cursor.execute, "select :1 from dual",
                                   (in_value,))

    def test_2229_return_float_from_division(self):
        "2229 - test that fetching the result of division returns a float"
        self.cursor.execute("""
                select IntCol / 7
                from TestNumbers
                where IntCol = 1""")
        result, = self.cursor.fetchone()
        self.assertEqual(result, 1.0 / 7.0)
        self.assertTrue(isinstance(result, float), "float not returned")

    def test_2230_string_format(self):
        "2230 - test that string format is returned properly"
        var = self.cursor.var(oracledb.NUMBER)
        self.assertIs(var.type, oracledb.DB_TYPE_NUMBER)
        self.assertEqual(str(var),
                "<oracledb.Var of type DB_TYPE_NUMBER with value None>")
        var.setvalue(0, 4.5)
        self.assertEqual(str(var),
                "<oracledb.Var of type DB_TYPE_NUMBER with value 4.5>")

    def test_2231_bind_binary_double(self):
        "2231 - test that binding binary double is possible"
        statement = "select :1 from dual"
        self.cursor.setinputsizes(oracledb.DB_TYPE_BINARY_DOUBLE)
        self.cursor.execute(statement, (5,))
        self.assertEqual(self.cursor.bindvars[0].type,
                         oracledb.DB_TYPE_BINARY_DOUBLE)
        value, = self.cursor.fetchone()
        self.assertEqual(value, 5)
        self.cursor.execute(statement, (1.5,))
        self.assertEqual(self.cursor.bindvars[0].type,
                         oracledb.DB_TYPE_BINARY_DOUBLE)
        value, = self.cursor.fetchone()
        self.assertEqual(value, 1.5)
        self.cursor.execute(statement, (decimal.Decimal("NaN"),))
        self.assertEqual(self.cursor.bindvars[0].type,
                         oracledb.DB_TYPE_BINARY_DOUBLE)
        value, = self.cursor.fetchone()
        self.assertEqual(str(value), str(float("NaN")))

    def test_2232_fetch_binary_int(self):
        "2232 - test fetching numbers as binary integers"
        self.cursor.outputtypehandler = self.output_type_handler_binary_int
        for value in (1, 2 ** 31, 2 ** 63 - 1, -1, -2 ** 31, -2 ** 63 + 1):
            self.cursor.execute("select :1 from dual", [str(value)])
            fetched_value, = self.cursor.fetchone()
            self.assertEqual(value, fetched_value)

    def test_2233_out_bind_binary_int(self):
        "2233 - test binding native integer as an out bind"
        statement = "begin :value := 2.9; end;"
        simple_var = self.cursor.var(oracledb.DB_TYPE_BINARY_INTEGER)
        self.cursor.execute(statement, [simple_var])
        self.assertEqual(simple_var.getvalue(), 2)

        statement = "begin :value := 1.5; end;"
        simple_var = self.cursor.var(oracledb.DB_TYPE_BINARY_INTEGER)
        self.cursor.execute(statement, [simple_var])
        self.assertEqual(simple_var.getvalue(), 1)

    def test_2234_in_bind_binary_int(self):
        "2234 - test binding in a native integer"
        statement = "begin :value := :value + 2.5; end;"
        simple_var = self.cursor.var(oracledb.DB_TYPE_BINARY_INTEGER)
        simple_var.setvalue(0, 0)
        self.cursor.execute(statement, [simple_var])
        self.assertEqual(simple_var.getvalue(), 2)

        simple_var.setvalue(0, -5)
        self.cursor.execute(statement, [simple_var])
        self.assertEqual(simple_var.getvalue(), -2)

    def test_2235_setting_decimal_value_binary_int(self):
        "2235 - test setting decimal value for binary int"
        statement = "begin :value := :value + 2.5; end;"
        simple_var = self.cursor.var(oracledb.DB_TYPE_BINARY_INTEGER)
        simple_var.setvalue(0, 2.5)
        self.cursor.execute(statement, [simple_var])
        self.assertEqual(simple_var.getvalue(), 4)

    def test_2236_out_bind_binary_int_with_large_value(self):
        "2236 - bind a large value to binary int"
        statement = "begin :value := POWER(2, 31) - 1; end;"
        simple_var = self.cursor.var(oracledb.DB_TYPE_BINARY_INTEGER)
        self.cursor.execute(statement, [simple_var])
        self.assertEqual(simple_var.getvalue(), 2**31 - 1)

        statement = "begin :value := POWER(-2, 31) - 1; end;"
        self.cursor.execute(statement, [simple_var])
        self.assertEqual(simple_var.getvalue(), -2**31 - 1)

    def test_2237_fetch_number_with_lobs_default_false(self):
        "2237 - fetch a number with oracledb.defaults.fetch_lobs = False"
        with test_env.FetchLobsContextManager(False):
            self.cursor.execute("select 1 from dual")
            result, = self.cursor.fetchone()
            self.assertEqual(type(result), int)

    def test_2238_fetch_small_constant_with_decimal_point(self):
        "2238 - fetch a small constant with a decimal point"
        self.cursor.outputtypehandler = self.output_type_handler_str
        self.cursor.execute("select 3 / 2 from dual")
        result, = self.cursor.fetchone()
        self.assertTrue(len(result) == 3 and result[0] == "1" \
                        and result[-1] == "5")

if __name__ == "__main__":
    test_env.run_test_cases()



#------------------------------------------------------------------------------
# Copyright (c) 2020, 2023, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# cursor.pyx
#
# Cython file defining the thin Cursor implementation class (embedded in
# thin_impl.pyx).
#------------------------------------------------------------------------------

cdef class ThinCursorImpl(BaseCursorImpl):

    cdef:
        ThinConnImpl _conn_impl
        Statement _statement
        list _batcherrors
        list _dmlrowcounts
        list _implicit_resultsets
        uint32_t _num_columns
        uint32_t _last_row_index
        Rowid _lastrowid

    def __cinit__(self, conn_impl):
        self._conn_impl = conn_impl

    cdef int _close(self, bint in_del) except -1:
        if self._statement is not None:
            self._conn_impl._return_statement(self._statement)
            self._statement = None

    cdef BaseVarImpl _create_var_impl(self, object conn):
        cdef ThinVarImpl var_impl
        var_impl = ThinVarImpl.__new__(ThinVarImpl)
        var_impl._conn_impl = self._conn_impl
        return var_impl

    cdef MessageWithData _create_message(self, type typ, object cursor):
        """
        Creates a message object that is used to send a request to the database
        and receive back its response.
        """
        cdef MessageWithData message
        message = typ.__new__(typ, cursor, self)
        message._initialize(self._conn_impl)
        message.cursor = cursor
        message.cursor_impl = self
        return message

    @cython.boundscheck(False)
    @cython.wraparound(False)
    cdef int _create_fetch_var(self, object conn, object cursor,
                               object type_handler, ssize_t pos,
                               FetchInfo fetch_info) except -1:
        """
        Internal method that creates a fetch variable. A check is made after
        the variable is created to determine if a conversion is required and
        therefore a define must be performed.
        """
        cdef:
            ThinDbObjectTypeImpl typ_impl
            ThinVarImpl var_impl
        BaseCursorImpl._create_fetch_var(self, conn, cursor, type_handler, pos,
                                         fetch_info)
        var_impl = self.fetch_var_impls[pos]
        if var_impl.dbtype._ora_type_num != fetch_info._dbtype._ora_type_num:
            conversion_helper(var_impl, fetch_info,
                              &self._statement._requires_define)
        elif not self._statement._requires_define \
                and fetch_info._dbtype._ora_type_num in \
                        (TNS_DATA_TYPE_BLOB, TNS_DATA_TYPE_CLOB,
                         TNS_DATA_TYPE_JSON):
            self._statement._requires_define = True
        elif var_impl.objtype is not None:
            typ_impl = var_impl.objtype
            if typ_impl.is_xml_type:
                var_impl.outconverter = \
                        lambda v: v if isinstance(v, str) else v.read()

    cdef int _fetch_rows(self, object cursor) except -1:
        """
        Internal method used for fetching rows from the database.
        """
        cdef MessageWithData message
        if self._statement._requires_full_execute:
            message = self._create_message(ExecuteMessage, cursor)
        else:
            message = self._create_message(FetchMessage, cursor)
        self._conn_impl._protocol._process_single_message(message)
        self._statement._requires_full_execute = False

    cdef BaseConnImpl _get_conn_impl(self):
        """
        Internal method used to return the connection implementation associated
        with the cursor implementation.
        """
        return self._conn_impl

    cdef bint _is_plsql(self):
        return self._statement._is_plsql

    cdef int _preprocess_execute(self, object conn) except -1:
        cdef BindInfo bind_info
        if self.bind_vars is not None:
            self._perform_binds(conn, 0)
        for bind_info in self._statement._bind_info_list:
            if bind_info._bind_var_impl is None:
                errors._raise_err(errors.ERR_MISSING_BIND_VALUE,
                                  name=bind_info._bind_name)

    def execute(self, cursor):
        cdef:
            object conn = cursor.connection
            MessageWithData message
        self._preprocess_execute(conn)
        message = self._create_message(ExecuteMessage, cursor)
        message.num_execs = 1
        self._conn_impl._protocol._process_single_message(message)
        self._statement._requires_full_execute = False
        if self._statement._is_query:
            self.rowcount = 0
            if message.type_cache is not None:
                message.type_cache.populate_partial_types(conn)

    def executemany(self, cursor, num_execs, batcherrors, arraydmlrowcounts):
        cdef:
            MessageWithData messsage
            Statement stmt
            uint32_t i

        # set up message to send
        self._preprocess_execute(cursor.connection)
        message = self._create_message(ExecuteMessage, cursor)
        message.num_execs = num_execs
        message.batcherrors = batcherrors
        message.arraydmlrowcounts = arraydmlrowcounts
        stmt = self._statement

        # only DML statements may use the batch errors or array DML row counts
        # flags
        if not stmt._is_dml and (batcherrors or arraydmlrowcounts):
            errors._raise_err(errors.ERR_EXECUTE_MODE_ONLY_FOR_DML)

        # if a PL/SQL statement requires a full execute, perform only a single
        # iteration in order to allow the determination of input/output binds
        # to be completed; after that, an execution of the remaining iterations
        # can be performed
        if stmt._is_plsql and stmt._requires_full_execute:
            message.num_execs = 1
            self._conn_impl._protocol._process_single_message(message)
            stmt._requires_full_execute = False
            if num_execs == 1:
                return
            message.offset = 1
            message.num_execs = num_execs - 1
        self._conn_impl._protocol._process_single_message(message)
        stmt._requires_full_execute = False

    def get_array_dml_row_counts(self):
        if self._dmlrowcounts is None:
            errors._raise_err(errors.ERR_ARRAY_DML_ROW_COUNTS_NOT_ENABLED)
        return self._dmlrowcounts

    def get_batch_errors(self):
        return self._batcherrors

    def get_bind_names(self):
        return list(self._statement._bind_info_dict.keys())

    def get_implicit_results(self, connection):
        if self._implicit_resultsets is None:
            errors._raise_err(errors.ERR_NO_STATEMENT_EXECUTED)
        return self._implicit_resultsets

    def get_lastrowid(self):
        if self.rowcount > 0:
            return _encode_rowid(&self._lastrowid)

    def is_query(self, connection):
        return self.fetch_vars is not None

    def parse(self, cursor):
        cdef MessageWithData message
        message = self._create_message(ExecuteMessage, cursor)
        message.parse_only = True
        self._conn_impl._protocol._process_single_message(message)

    def prepare(self, str sql, str tag, bint cache_statement):
        self.statement = sql
        if self._statement is not None:
            self._conn_impl._return_statement(self._statement)
            self._statement = None
        self._statement = self._conn_impl._get_statement(sql.strip(),
                                                         cache_statement)
        self.fetch_vars = self._statement._fetch_vars
        self.fetch_var_impls = self._statement._fetch_var_impls
        self._num_columns = self._statement._num_columns
